@@ -9,7 +9,7 @@
 // the LICENSE-APACHE file) or the MIT license (found in
 // the LICENSE-MIT file), at your option.
 
-use std::{
+use core::{
     fmt,
     ops::{Add, AddAssign, Div, DivAssign, Mul, MulAssign, Neg, Sub, SubAssign},
 };
@@ -48,7 +48,7 @@ impl Affine {
     /// [Wikipedia](https://en.wikipedia.org/wiki/Affine_transformation)
     /// formulation of affine transformation as augmented matrix. The
     /// idea is that `(A * B) * v == A * (B * v)`, where `*` is the
-    /// [`Mul`](std::ops::Mul) trait.
+    /// [`Mul`](core::ops::Mul) trait.
     #[inline]
     pub const fn new(c: [f64; 6]) -> Affine {
         Affine(c)
@@ -67,20 +67,6 @@ impl Affine {
         Affine([s_x, 0.0, 0.0, s_y, 0.0, 0.0])
     }
 
-    /// An affine transform representing rotation.
-    ///
-    /// The convention for rotation is that a positive angle rotates a
-    /// positive X direction into positive Y. Thus, in a Y-down coordinate
-    /// system (as is common for graphics), it is a clockwise rotation, and
-    /// in Y-up (traditional for math), it is anti-clockwise.
-    ///
-    /// The angle, `th`, is expressed in radians.
-    #[inline]
-    pub fn rotate(th: f64) -> Affine {
-        let (s, c) = th.sin_cos();
-        Affine([c, s, -s, c, 0.0, 0.0])
-    }
-
     /// An affine transform representing translation.
     #[inline]
     pub fn translate<V: Into<Vec2>>(p: V) -> Affine {
