@@ -11,13 +11,13 @@ use std::{cell::RefCell, num::NonZeroU32, rc::Rc, sync::Arc, time::Instant};
 
 use egui_winit::ActionRequested;
 use glutin::{
-    config::GlConfig,
-    context::NotCurrentGlContext,
-    display::GetGlDisplay,
-    prelude::{GlDisplay, PossiblyCurrentGlContext},
-    surface::GlSurface,
+    config::GlConfig as _,
+    context::NotCurrentGlContext as _,
+    display::GetGlDisplay as _,
+    prelude::{GlDisplay as _, PossiblyCurrentGlContext as _},
+    surface::GlSurface as _,
 };
-use raw_window_handle::HasWindowHandle;
+use raw_window_handle::HasWindowHandle as _;
 use winit::{
     event_loop::{ActiveEventLoop, EventLoop, EventLoopProxy},
     window::{Window, WindowId},
@@ -32,13 +32,13 @@ use egui::{
 use egui_winit::accesskit_winit;
 
 use crate::{
-    native::epi_integration::EpiIntegration, App, AppCreator, CreationContext, NativeOptions,
-    Result, Storage,
+    App, AppCreator, CreationContext, NativeOptions, Result, Storage,
+    native::epi_integration::EpiIntegration,
 };
 
 use super::{
     epi_integration, event_loop_context,
-    winit_integration::{create_egui_context, EventResult, UserEvent, WinitApp},
+    winit_integration::{EventResult, UserEvent, WinitApp, create_egui_context},
 };
 
 // ----------------------------------------------------------------------------
@@ -129,7 +129,7 @@ impl<'app> GlowWinitApp<'app> {
         native_options: NativeOptions,
         app_creator: AppCreator<'app>,
     ) -> Self {
-        crate::profile_function!();
+        profiling::function_scope!();
         Self {
             repaint_proxy: Arc::new(egui::mutex::Mutex::new(event_loop.create_proxy())),
             app_name: app_name.to_owned(),
@@ -139,15 +139,14 @@ impl<'app> GlowWinitApp<'app> {
         }
     }
 
-    #[allow(unsafe_code)]
+    #[expect(unsafe_code)]
     fn create_glutin_windowed_context(
         egui_ctx: &egui::Context,
         event_loop: &ActiveEventLoop,
         storage: Option<&dyn Storage>,
         native_options: &mut NativeOptions,
     ) -> Result<(GlutinWindowContext, egui_glow::Painter)> {
-        crate::profile_function!();
-
+        profiling::function_scope!();
         let window_settings = epi_integration::load_window_settings(storage);
 
         let winit_window_builder = epi_integration::viewport_builder(
@@ -172,7 +171,7 @@ impl<'app> GlowWinitApp<'app> {
         }
 
         let gl = unsafe {
-            crate::profile_scope!("glow::Context::from_loader_function");
+            profiling::scope!("glow::Context::from_loader_function");
             Arc::new(glow::Context::from_loader_function(|s| {
                 let s = std::ffi::CString::new(s)
                     .expect("failed to construct C string from string for gl proc address");
@@ -195,7 +194,7 @@ impl<'app> GlowWinitApp<'app> {
         &mut self,
         event_loop: &ActiveEventLoop,
     ) -> Result<&mut GlowWinitRunning<'app>> {
-        crate::profile_function!();
+        profiling::function_scope!();
 
         let storage = if let Some(file) = &self.native_options.persistence_path {
             epi_integration::create_storage_with_file(file)
@@ -273,7 +272,7 @@ impl<'app> GlowWinitApp<'app> {
                 ..
             } = viewport
             {
-                egui_winit.init_accesskit(window, event_loop_proxy);
+                egui_winit.init_accesskit(event_loop, window, event_loop_proxy);
             }
         }
 
@@ -308,7 +307,7 @@ impl<'app> GlowWinitApp<'app> {
                 raw_display_handle: window.display_handle().map(|h| h.as_raw()),
                 raw_window_handle: window.window_handle().map(|h| h.as_raw()),
             };
-            crate::profile_scope!("app_creator");
+            profiling::scope!("app_creator");
             app_creator(&cc).map_err(crate::Error::AppCreation)?
         };
 
@@ -345,7 +344,7 @@ impl<'app> GlowWinitApp<'app> {
     }
 }
 
-impl<'app> WinitApp for GlowWinitApp<'app> {
+impl WinitApp for GlowWinitApp<'_> {
     fn egui_ctx(&self) -> Option<&egui::Context> {
         self.running.as_ref().map(|r| &r.integration.egui_ctx)
     }
@@ -367,9 +366,23 @@ impl<'app> WinitApp for GlowWinitApp<'app> {
             .and_then(|r| r.glutin.borrow().window_from_viewport.get(&id).copied())
     }
 
+    fn save(&mut self) {
+        log::debug!("WinitApp::save called");
+        if let Some(running) = self.running.as_mut() {
+            profiling::function_scope!();
+
+            // This is used because of the "save on suspend" logic on Android. Once the application is suspended, there is no window associated to it, which was causing panics when `.window().expect()` was used.
+            let window_opt = running.glutin.borrow().window_opt(ViewportId::ROOT);
+
+            running
+                .integration
+                .save(running.app.as_mut(), window_opt.as_deref());
+        }
+    }
+
     fn save_and_destroy(&mut self) {
         if let Some(mut running) = self.running.take() {
-            crate::profile_function!();
+            profiling::function_scope!();
 
             running.integration.save(
                 running.app.as_mut(),
@@ -414,7 +427,7 @@ impl<'app> WinitApp for GlowWinitApp<'app> {
         if let Some(running) = &mut self.running {
             running.glutin.borrow_mut().on_suspend()?;
         }
-        Ok(EventResult::Wait)
+        Ok(EventResult::Save)
     }
 
     fn device_event(
@@ -480,13 +493,13 @@ impl<'app> WinitApp for GlowWinitApp<'app> {
     }
 }
 
-impl<'app> GlowWinitRunning<'app> {
+impl GlowWinitRunning<'_> {
     fn run_ui_and_paint(
         &mut self,
         event_loop: &ActiveEventLoop,
         window_id: WindowId,
     ) -> Result<EventResult> {
-        crate::profile_function!();
+        profiling::function_scope!();
 
         let Some(viewport_id) = self
             .glutin
@@ -498,8 +511,7 @@ impl<'app> GlowWinitRunning<'app> {
             return Ok(EventResult::Wait);
         };
 
-        #[cfg(feature = "puffin")]
-        puffin::GlobalProfiler::lock().new_frame();
+        profiling::finish_frame!();
 
         let mut frame_timer = crate::stopwatch::Stopwatch::new();
         frame_timer.start();
@@ -661,13 +673,14 @@ impl<'app> GlowWinitRunning<'app> {
         {
             for action in viewport.actions_requested.drain() {
                 match action {
-                    ActionRequested::Screenshot => {
+                    ActionRequested::Screenshot(user_data) => {
                         let screenshot = painter.read_screen_rgba(screen_size_in_pixels);
                         egui_winit
                             .egui_input_mut()
                             .events
                             .push(egui::Event::Screenshot {
                                 viewport_id,
+                                user_data,
                                 image: screenshot.into(),
                             });
                     }
@@ -697,7 +710,7 @@ impl<'app> GlowWinitRunning<'app> {
         {
             // vsync - don't count as frame-time:
             frame_timer.pause();
-            crate::profile_scope!("swap_buffers");
+            profiling::scope!("swap_buffers");
             let context = current_gl_context
                 .as_ref()
                 .ok_or(egui_glow::PainterError::from(
@@ -725,7 +738,7 @@ impl<'app> GlowWinitRunning<'app> {
         if window.is_minimized() == Some(true) {
             // On Mac, a minimized Window uses up all CPU:
             // https://github.com/emilk/egui/issues/325
-            crate::profile_scope!("minimized_sleep");
+            profiling::scope!("minimized_sleep");
             std::thread::sleep(std::time::Duration::from_millis(10));
         }
 
@@ -856,7 +869,7 @@ fn change_gl_context(
     not_current_gl_context: &mut Option<glutin::context::NotCurrentContext>,
     gl_surface: &glutin::surface::Surface<glutin::surface::WindowSurface>,
 ) {
-    crate::profile_function!();
+    profiling::function_scope!();
 
     if !cfg!(target_os = "windows") {
         // According to https://github.com/emilk/egui/issues/4289
@@ -865,7 +878,7 @@ fn change_gl_context(
         // See https://github.com/emilk/egui/issues/4173
 
         if let Some(current_gl_context) = current_gl_context {
-            crate::profile_scope!("is_current");
+            profiling::scope!("is_current");
             if gl_surface.is_current(current_gl_context) {
                 return; // Early-out to save a lot of time.
             }
@@ -875,7 +888,7 @@ fn change_gl_context(
     let not_current = if let Some(not_current_context) = not_current_gl_context.take() {
         not_current_context
     } else {
-        crate::profile_scope!("make_not_current");
+        profiling::scope!("make_not_current");
         current_gl_context
             .take()
             .unwrap()
@@ -883,19 +896,19 @@ fn change_gl_context(
             .unwrap()
     };
 
-    crate::profile_scope!("make_current");
+    profiling::scope!("make_current");
     *current_gl_context = Some(not_current.make_current(gl_surface).unwrap());
 }
 
 impl GlutinWindowContext {
-    #[allow(unsafe_code)]
+    #[expect(unsafe_code)]
     unsafe fn new(
         egui_ctx: &egui::Context,
         viewport_builder: ViewportBuilder,
         native_options: &NativeOptions,
         event_loop: &ActiveEventLoop,
     ) -> Result<Self> {
-        crate::profile_function!();
+        profiling::function_scope!();
 
         // There is a lot of complexity with opengl creation,
         // so prefer extensive logging to get all the help we can to debug issues.
@@ -941,15 +954,16 @@ impl GlutinWindowContext {
         // Create GL display. This may probably create a window too on most platforms. Definitely on `MS windows`. Never on Android.
         let display_builder = glutin_winit::DisplayBuilder::new()
             // we might want to expose this option to users in the future. maybe using an env var or using native_options.
-            .with_preference(glutin_winit::ApiPreference::FallbackEgl) // https://github.com/emilk/egui/issues/2520#issuecomment-1367841150
+            //
+            // The justification for FallbackEgl over PreferEgl is at https://github.com/emilk/egui/pull/2526#issuecomment-1400229576 .
+            .with_preference(glutin_winit::ApiPreference::FallbackEgl)
             .with_window_attributes(Some(egui_winit::create_winit_window_attributes(
                 egui_ctx,
-                event_loop,
                 viewport_builder.clone(),
             )));
 
         let (window, gl_config) = {
-            crate::profile_scope!("DisplayBuilder::build");
+            profiling::scope!("DisplayBuilder::build");
 
             display_builder
                 .build(
@@ -992,7 +1006,7 @@ impl GlutinWindowContext {
             .build(glutin_raw_window_handle);
 
         let gl_context_result = unsafe {
-            crate::profile_scope!("create_context");
+            profiling::scope!("create_context");
             gl_config
                 .display()
                 .create_context(&gl_config, &context_attributes)
@@ -1001,7 +1015,9 @@ impl GlutinWindowContext {
         let gl_context = match gl_context_result {
             Ok(it) => it,
             Err(err) => {
-                log::warn!("Failed to create context using default context attributes {context_attributes:?} due to error: {err}");
+                log::warn!(
+                    "Failed to create context using default context attributes {context_attributes:?} due to error: {err}"
+                );
                 log::debug!(
                     "Retrying with fallback context attributes: {fallback_context_attributes:?}"
                 );
@@ -1067,7 +1083,7 @@ impl GlutinWindowContext {
     ///
     /// Errors will be logged.
     fn initialize_all_windows(&mut self, event_loop: &ActiveEventLoop) {
-        crate::profile_function!();
+        profiling::function_scope!();
 
         let viewports: Vec<ViewportId> = self.viewports.keys().copied().collect();
 
@@ -1079,13 +1095,13 @@ impl GlutinWindowContext {
     }
 
     /// Create a surface, window, and winit integration for the viewport, if missing.
-    #[allow(unsafe_code)]
+    #[expect(unsafe_code)]
     pub(crate) fn initialize_window(
         &mut self,
         viewport_id: ViewportId,
         event_loop: &ActiveEventLoop,
     ) -> Result {
-        crate::profile_function!();
+        profiling::function_scope!();
 
         let viewport = self
             .viewports
@@ -1098,7 +1114,6 @@ impl GlutinWindowContext {
             log::debug!("Creating a window for viewport {viewport_id:?}");
             let window_attributes = egui_winit::create_winit_window_attributes(
                 &self.egui_ctx,
-                event_loop,
                 viewport.builder.clone(),
             );
             if window_attributes.transparent()
@@ -1213,10 +1228,12 @@ impl GlutinWindowContext {
             .expect("viewport doesn't exist")
     }
 
+    fn window_opt(&self, viewport_id: ViewportId) -> Option<Arc<Window>> {
+        self.viewport(viewport_id).window.clone()
+    }
+
     fn window(&self, viewport_id: ViewportId) -> Arc<Window> {
-        self.viewport(viewport_id)
-            .window
-            .clone()
+        self.window_opt(viewport_id)
             .expect("winit window doesn't exist")
     }
 
@@ -1265,7 +1282,7 @@ impl GlutinWindowContext {
         egui_ctx: &egui::Context,
         viewport_output: &ViewportIdMap<ViewportOutput>,
     ) {
-        crate::profile_function!();
+        profiling::function_scope!();
 
         for (
             viewport_id,
@@ -1326,7 +1343,7 @@ fn initialize_or_update_viewport(
     mut builder: ViewportBuilder,
     viewport_ui_cb: Option<Arc<dyn Fn(&egui::Context) + Send + Sync>>,
 ) -> &mut Viewport {
-    crate::profile_function!();
+    profiling::function_scope!();
 
     if builder.icon.is_none() {
         // Inherit icon from parent
@@ -1390,7 +1407,7 @@ fn render_immediate_viewport(
     beginning: Instant,
     immediate_viewport: ImmediateViewport<'_>,
 ) {
-    crate::profile_function!();
+    profiling::function_scope!();
 
     let ImmediateViewport {
         ids,
@@ -1513,7 +1530,7 @@ fn render_immediate_viewport(
     );
 
     {
-        crate::profile_scope!("swap_buffers");
+        profiling::scope!("swap_buffers");
         if let Err(err) = gl_surface.swap_buffers(current_gl_context) {
             log::error!("swap_buffers failed: {err}");
         }
@@ -1549,6 +1566,6 @@ fn save_screenshot_and_exit(
     });
     log::info!("Screenshot saved to {path:?}.");
 
-    #[allow(clippy::exit)]
+    #[expect(clippy::exit)]
     std::process::exit(0);
 }
