@@ -15,9 +15,11 @@
 //!
 //! Add your [`crate::Window`]:s after any top-level panels.
 
+use emath::GuiRounding as _;
+
 use crate::{
-    lerp, vec2, Align, Context, CursorIcon, Frame, Id, InnerResponse, LayerId, Layout, NumExt,
-    Rangef, Rect, Sense, Stroke, Ui, UiBuilder, UiKind, UiStackInfo, Vec2,
+    Align, Context, CursorIcon, Frame, Id, InnerResponse, LayerId, Layout, NumExt as _, Rangef,
+    Rect, Sense, Stroke, Ui, UiBuilder, UiKind, UiStackInfo, Vec2, lerp, vec2,
 };
 
 fn animate_expansion(ctx: &Context, id: Id, is_expanded: bool) -> f32 {
@@ -76,6 +78,13 @@ impl Side {
             Self::Right => rect.right(),
         }
     }
+
+    fn sign(self) -> f32 {
+        match self {
+            Self::Left => -1.0,
+            Self::Right => 1.0,
+        }
+    }
 }
 
 /// A panel that covers the entire left or right side of a [`Ui`] or screen.
@@ -264,6 +273,8 @@ impl SidePanel {
             }
         }
 
+        panel_rect = panel_rect.round_ui();
+
         let mut panel_ui = ui.new_child(
             UiBuilder::new()
                 .id_salt(id)
@@ -345,12 +356,8 @@ impl SidePanel {
             // TODO(emilk): draw line on top of all panels in this ui when https://github.com/emilk/egui/issues/1516 is done
             let resize_x = side.opposite().side_x(rect);
 
-            // This makes it pixel-perfect for odd-sized strokes (width=1.0, width=3.0, etc)
-            let resize_x = ui.painter().round_to_pixel_center(resize_x);
-
-            // We want the line exactly on the last pixel but rust rounds away from zero so we bring it back a bit for
-            // left-side panels
-            let resize_x = resize_x - if side == Side::Left { 1.0 } else { 0.0 };
+            // Make sure the line is on the inside of the panel:
+            let resize_x = resize_x + 0.5 * side.sign() * stroke.width;
             ui.painter().vline(resize_x, panel_rect.y_range(), stroke);
         }
 
@@ -372,14 +379,14 @@ impl SidePanel {
         ctx: &Context,
         add_contents: Box<dyn FnOnce(&mut Ui) -> R + 'c>,
     ) -> InnerResponse<R> {
-        let layer_id = LayerId::background();
         let side = self.side;
         let available_rect = ctx.available_rect();
         let mut panel_ui = Ui::new(
             ctx.clone(),
-            layer_id,
             self.id,
-            UiBuilder::new().max_rect(available_rect),
+            UiBuilder::new()
+                .layer_id(LayerId::background())
+                .max_rect(available_rect),
         );
         panel_ui.set_clip_rect(ctx.screen_rect());
 
@@ -558,6 +565,13 @@ impl TopBottomSide {
             Self::Bottom => rect.bottom(),
         }
     }
+
+    fn sign(self) -> f32 {
+        match self {
+            Self::Top => -1.0,
+            Self::Bottom => 1.0,
+        }
+    }
 }
 
 /// A panel that covers the entire top or bottom of a [`Ui`] or screen.
@@ -756,6 +770,8 @@ impl TopBottomPanel {
             }
         }
 
+        panel_rect = panel_rect.round_ui();
+
         let mut panel_ui = ui.new_child(
             UiBuilder::new()
                 .id_salt(id)
@@ -837,12 +853,8 @@ impl TopBottomPanel {
             // TODO(emilk): draw line on top of all panels in this ui when https://github.com/emilk/egui/issues/1516 is done
             let resize_y = side.opposite().side_y(rect);
 
-            // This makes it pixel-perfect for odd-sized strokes (width=1.0, width=3.0, etc)
-            let resize_y = ui.painter().round_to_pixel_center(resize_y);
-
-            // We want the line exactly on the last pixel but rust rounds away from zero so we bring it back a bit for
-            // top-side panels
-            let resize_y = resize_y - if side == TopBottomSide::Top { 1.0 } else { 0.0 };
+            // Make sure the line is on the inside of the panel:
+            let resize_y = resize_y + 0.5 * side.sign() * stroke.width;
             ui.painter().hline(panel_rect.x_range(), resize_y, stroke);
         }
 
@@ -864,15 +876,15 @@ impl TopBottomPanel {
         ctx: &Context,
         add_contents: Box<dyn FnOnce(&mut Ui) -> R + 'c>,
     ) -> InnerResponse<R> {
-        let layer_id = LayerId::background();
         let available_rect = ctx.available_rect();
         let side = self.side;
 
         let mut panel_ui = Ui::new(
             ctx.clone(),
-            layer_id,
             self.id,
-            UiBuilder::new().max_rect(available_rect),
+            UiBuilder::new()
+                .layer_id(LayerId::background())
+                .max_rect(available_rect),
         );
         panel_ui.set_clip_rect(ctx.screen_rect());
 
@@ -1130,15 +1142,14 @@ impl CentralPanel {
         ctx: &Context,
         add_contents: Box<dyn FnOnce(&mut Ui) -> R + 'c>,
     ) -> InnerResponse<R> {
-        let available_rect = ctx.available_rect();
-        let layer_id = LayerId::background();
         let id = Id::new((ctx.viewport_id(), "central_panel"));
 
         let mut panel_ui = Ui::new(
             ctx.clone(),
-            layer_id,
             id,
-            UiBuilder::new().max_rect(available_rect),
+            UiBuilder::new()
+                .layer_id(LayerId::background())
+                .max_rect(ctx.available_rect().round_ui()),
         );
         panel_ui.set_clip_rect(ctx.screen_rect());
 
