@@ -2,7 +2,7 @@ use ahash::HashMap;
 
 use emath::TSTransform;
 
-use crate::{ahash, emath, LayerId, Pos2, WidgetRect, WidgetRects};
+use crate::{LayerId, Pos2, Rect, Sense, WidgetRect, WidgetRects, ahash, emath, id::IdSet};
 
 /// Result of a hit-test against [`WidgetRects`].
 ///
@@ -12,11 +12,18 @@ use crate::{ahash, emath, LayerId, Pos2, WidgetRect, WidgetRects};
 /// or if we're currently already dragging something.
 #[derive(Clone, Debug, Default)]
 pub struct WidgetHits {
+    /// All widgets close to the pointer, back-to-front.
+    ///
+    /// This is a superset of all other widgets in this struct.
+    pub close: Vec<WidgetRect>,
+
     /// All widgets that contains the pointer, back-to-front.
     ///
-    /// i.e. both a Window and the button in it can contain the pointer.
+    /// i.e. both a Window and the Button in it can contain the pointer.
     ///
     /// Some of these may be widgets in a layer below the top-most layer.
+    ///
+    /// This will be used for hovering.
     pub contains_pointer: Vec<WidgetRect>,
 
     /// If the user would start a clicking now, this is what would be clicked.
@@ -35,18 +42,18 @@ pub struct WidgetHits {
 pub fn hit_test(
     widgets: &WidgetRects,
     layer_order: &[LayerId],
-    layer_transforms: &HashMap<LayerId, TSTransform>,
+    layer_to_global: &HashMap<LayerId, TSTransform>,
     pos: Pos2,
     search_radius: f32,
 ) -> WidgetHits {
-    crate::profile_function!();
+    profiling::function_scope!();
 
     let search_radius_sq = search_radius * search_radius;
 
     // Transform the position into the local coordinate space of each layer:
-    let pos_in_layers: HashMap<LayerId, Pos2> = layer_transforms
+    let pos_in_layers: HashMap<LayerId, Pos2> = layer_to_global
         .iter()
-        .map(|(layer_id, t)| (*layer_id, t.inverse() * pos))
+        .map(|(layer_id, to_global)| (*layer_id, to_global.inverse() * pos))
         .collect();
 
     let mut closest_dist_sq = f32::INFINITY;
@@ -58,11 +65,12 @@ pub fn hit_test(
         .filter(|layer| layer.order.allow_interaction())
         .flat_map(|&layer_id| widgets.get_layer(layer_id))
         .filter(|&w| {
-            if w.interact_rect.is_negative() {
+            if w.interact_rect.is_negative() || w.interact_rect.any_nan() {
                 return false;
             }
 
             let pos_in_layer = pos_in_layers.get(&w.layer_id).copied().unwrap_or(pos);
+            // TODO(emilk): we should probably do the distance testing in global space instead
             let dist_sq = w.interact_rect.distance_sq_to_pos(pos_in_layer);
 
             // In tie, pick last = topmost.
@@ -76,51 +84,136 @@ pub fn hit_test(
         .copied()
         .collect();
 
-    // We need to pick one single layer for the interaction.
-    if let Some(closest_hit) = closest_hit {
-        // Select the top layer, and ignore widgets in any other layer:
-        let top_layer = closest_hit.layer_id;
-        close.retain(|w| w.layer_id == top_layer);
+    // Transform to global coordinates:
+    for hit in &mut close {
+        if let Some(to_global) = layer_to_global.get(&hit.layer_id).copied() {
+            *hit = hit.transform(to_global);
+        }
+    }
 
-        // If the widget is disabled, treat it as if it isn't sensing anything.
-        // This simplifies the code in `hit_test_on_close` so it doesn't have to check
-        // the `enabled` flag everywhere:
-        for w in &mut close {
-            if !w.enabled {
-                w.sense.click = false;
-                w.sense.drag = false;
+    close.retain(|rect| !rect.interact_rect.any_nan()); // Protect against bad input and transforms
+
+    // When using layer transforms it is common to stack layers close to each other.
+    // For instance, you may have a resize-separator on a panel, with two
+    // transform-layers on either side.
+    // The resize-separator is technically in a layer _behind_ the transform-layers,
+    // but the user doesn't perceive it as such.
+    // So how do we handle this case?
+    //
+    // If we just allow interactions with ALL close widgets,
+    // then we might accidentally allow clicks through windows and other bad stuff.
+    //
+    // Let's try this:
+    // * Set up a hit-area (based on search_radius)
+    // * Iterate over all hits top-to-bottom
+    //   * Stop if any hit covers the whole hit-area, otherwise keep going
+    //   * Collect the layers ids in a set
+    // * Remove all widgets not in the above layer set
+    //
+    // This will most often result in only one layer,
+    // but if the pointer is at the edge of a layer, we might include widgets in
+    // a layer behind it.
+
+    let mut included_layers: ahash::HashSet<LayerId> = Default::default();
+    for hit in close.iter().rev() {
+        included_layers.insert(hit.layer_id);
+        let hit_covers_search_area = contains_circle(hit.interact_rect, pos, search_radius);
+        if hit_covers_search_area {
+            break; // nothing behind this layer could ever be interacted with
+        }
+    }
+
+    close.retain(|hit| included_layers.contains(&hit.layer_id));
+
+    // If a widget is disabled, treat it as if it isn't sensing anything.
+    // This simplifies the code in `hit_test_on_close` so it doesn't have to check
+    // the `enabled` flag everywhere:
+    for w in &mut close {
+        if !w.enabled {
+            w.sense -= Sense::CLICK;
+            w.sense -= Sense::DRAG;
+        }
+    }
+
+    // Find widgets which are hidden behind another widget and discard them.
+    // This is the case when a widget fully contains another widget and is on a different layer.
+    // It prevents "hovering through" widgets when there is a clickable widget behind.
+
+    let mut hidden = IdSet::default();
+    for (i, current) in close.iter().enumerate().rev() {
+        for next in &close[i + 1..] {
+            if next.interact_rect.contains_rect(current.interact_rect)
+                && current.layer_id != next.layer_id
+            {
+                hidden.insert(current.id);
             }
         }
-
-        let pos_in_layer = pos_in_layers.get(&top_layer).copied().unwrap_or(pos);
-        let hits = hit_test_on_close(&close, pos_in_layer);
-
-        if let Some(drag) = hits.drag {
-            debug_assert!(drag.sense.drag);
-        }
-        if let Some(click) = hits.click {
-            debug_assert!(click.sense.click);
-        }
-
-        hits
-    } else {
-        // No close widgets.
-        Default::default()
     }
-}
 
-fn hit_test_on_close(close: &[WidgetRect], pos: Pos2) -> WidgetHits {
-    #![allow(clippy::collapsible_else_if)]
+    close.retain(|c| !hidden.contains(&c.id));
 
-    // Only those widgets directly under the `pos`.
-    let hits: Vec<WidgetRect> = close
+    let mut hits = hit_test_on_close(&close, pos);
+
+    hits.contains_pointer = close
         .iter()
         .filter(|widget| widget.interact_rect.contains(pos))
         .copied()
         .collect();
 
-    let hit_click = hits.iter().copied().filter(|w| w.sense.click).last();
-    let hit_drag = hits.iter().copied().filter(|w| w.sense.drag).last();
+    hits.close = close;
+
+    {
+        // Undo the to_global-transform we applied earlier,
+        // go back to local layer-coordinates:
+
+        let restore_widget_rect = |w: &mut WidgetRect| {
+            *w = widgets.get(w.id).copied().unwrap_or(*w);
+        };
+
+        for wr in &mut hits.close {
+            restore_widget_rect(wr);
+        }
+        for wr in &mut hits.contains_pointer {
+            restore_widget_rect(wr);
+        }
+        if let Some(wr) = &mut hits.drag {
+            debug_assert!(
+                wr.sense.senses_drag(),
+                "We should only return drag hits if they sense drag"
+            );
+            restore_widget_rect(wr);
+        }
+        if let Some(wr) = &mut hits.click {
+            debug_assert!(
+                wr.sense.senses_click(),
+                "We should only return click hits if they sense click"
+            );
+            restore_widget_rect(wr);
+        }
+    }
+
+    hits
+}
+
+/// Returns true if the rectangle contains the whole circle.
+fn contains_circle(interact_rect: emath::Rect, pos: Pos2, radius: f32) -> bool {
+    interact_rect.shrink(radius).contains(pos)
+}
+
+fn hit_test_on_close(close: &[WidgetRect], pos: Pos2) -> WidgetHits {
+    #![allow(clippy::collapsible_else_if)]
+
+    // First find the best direct hits:
+    let hit_click = find_closest_within(
+        close.iter().copied().filter(|w| w.sense.senses_click()),
+        pos,
+        0.0,
+    );
+    let hit_drag = find_closest_within(
+        close.iter().copied().filter(|w| w.sense.senses_drag()),
+        pos,
+        0.0,
+    );
 
     match (hit_click, hit_drag) {
         (None, None) => {
@@ -130,22 +223,22 @@ fn hit_test_on_close(close: &[WidgetRect], pos: Pos2) -> WidgetHits {
                 close
                     .iter()
                     .copied()
-                    .filter(|w| w.sense.click || w.sense.drag),
+                    .filter(|w| w.sense.senses_click() || w.sense.senses_drag()),
                 pos,
             );
 
             if let Some(closest) = closest {
                 WidgetHits {
-                    contains_pointer: hits,
-                    click: closest.sense.click.then_some(closest),
-                    drag: closest.sense.drag.then_some(closest),
+                    click: closest.sense.senses_click().then_some(closest),
+                    drag: closest.sense.senses_drag().then_some(closest),
+                    ..Default::default()
                 }
             } else {
                 // Found nothing
                 WidgetHits {
-                    contains_pointer: hits,
                     click: None,
                     drag: None,
+                    ..Default::default()
                 }
             }
         }
@@ -158,9 +251,12 @@ fn hit_test_on_close(close: &[WidgetRect], pos: Pos2) -> WidgetHits {
             // or a moveable window.
             // It could also be something small, like a slider, or panel resize handle.
 
-            let closest_click = find_closest(close.iter().copied().filter(|w| w.sense.click), pos);
+            let closest_click = find_closest(
+                close.iter().copied().filter(|w| w.sense.senses_click()),
+                pos,
+            );
             if let Some(closest_click) = closest_click {
-                if closest_click.sense.drag {
+                if closest_click.sense.senses_drag() {
                     // We have something close that sense both clicks and drag.
                     // Should we use it over the direct drag-hit?
                     if hit_drag
@@ -170,21 +266,21 @@ fn hit_test_on_close(close: &[WidgetRect], pos: Pos2) -> WidgetHits {
                         // This is a smaller thing on a big background - help the user hit it,
                         // and ignore the big drag background.
                         WidgetHits {
-                            contains_pointer: hits,
                             click: Some(closest_click),
                             drag: Some(closest_click),
+                            ..Default::default()
                         }
                     } else {
-                        // The drag wiudth is separate from the click wiudth,
-                        // so return only the drag widget
+                        // The drag-widget is separate from the click-widget,
+                        // so return only the drag-widget
                         WidgetHits {
-                            contains_pointer: hits,
                             click: None,
                             drag: Some(hit_drag),
+                            ..Default::default()
                         }
                     }
                 } else {
-                    // These is a close pure-click widget.
+                    // This is a close pure-click widget.
                     // However, we should be careful to only return two different widgets
                     // when it is absolutely not going to confuse the user.
                     if hit_drag
@@ -194,17 +290,17 @@ fn hit_test_on_close(close: &[WidgetRect], pos: Pos2) -> WidgetHits {
                         // The drag widget is a big background thing (scroll area),
                         // so returning a separate click widget should not be confusing
                         WidgetHits {
-                            contains_pointer: hits,
                             click: Some(closest_click),
                             drag: Some(hit_drag),
+                            ..Default::default()
                         }
                     } else {
                         // The two widgets are just two normal small widgets close to each other.
                         // Highlighting both would be very confusing.
                         WidgetHits {
-                            contains_pointer: hits,
                             click: None,
                             drag: Some(hit_drag),
+                            ..Default::default()
                         }
                     }
                 }
@@ -217,7 +313,7 @@ fn hit_test_on_close(close: &[WidgetRect], pos: Pos2) -> WidgetHits {
                     close
                         .iter()
                         .copied()
-                        .filter(|w| w.sense.drag && w.id != hit_drag.id),
+                        .filter(|w| w.sense.senses_drag() && w.id != hit_drag.id),
                     pos,
                 );
 
@@ -229,17 +325,17 @@ fn hit_test_on_close(close: &[WidgetRect], pos: Pos2) -> WidgetHits {
                         // `hit_drag` is a big background thing and `closest_drag` is something small on top of it.
                         // Be helpful and return the small things:
                         return WidgetHits {
-                            contains_pointer: hits,
                             click: None,
                             drag: Some(closest_drag),
+                            ..Default::default()
                         };
                     }
                 }
 
                 WidgetHits {
-                    contains_pointer: hits,
                     click: None,
                     drag: Some(hit_drag),
+                    ..Default::default()
                 }
             }
         }
@@ -253,57 +349,57 @@ fn hit_test_on_close(close: &[WidgetRect], pos: Pos2) -> WidgetHits {
             // where when hovering directly over a drag-widget (like a big ScrollArea),
             // we look for close click-widgets (e.g. buttons).
             // This is because big background drag-widgets (ScrollArea, Window) are common,
-            // but bit clickable things aren't.
+            // but big clickable things aren't.
             // Even if they were, I think it would be confusing for a user if clicking
             // a drag-only widget would click something _behind_ it.
 
             WidgetHits {
-                contains_pointer: hits,
                 click: Some(hit_click),
                 drag: None,
+                ..Default::default()
             }
         }
 
         (Some(hit_click), Some(hit_drag)) => {
             // We have a perfect hit on both click and drag. Which is the topmost?
-            let click_idx = hits.iter().position(|w| *w == hit_click).unwrap();
-            let drag_idx = hits.iter().position(|w| *w == hit_drag).unwrap();
+            let click_idx = close.iter().position(|w| *w == hit_click).unwrap();
+            let drag_idx = close.iter().position(|w| *w == hit_drag).unwrap();
 
             let click_is_on_top_of_drag = drag_idx < click_idx;
             if click_is_on_top_of_drag {
-                if hit_click.sense.drag {
+                if hit_click.sense.senses_drag() {
                     // The top thing senses both clicks and drags.
                     WidgetHits {
-                        contains_pointer: hits,
                         click: Some(hit_click),
                         drag: Some(hit_click),
+                        ..Default::default()
                     }
                 } else {
                     // They are interested in different things,
                     // and click is on top. Report both hits,
                     // e.g. the top Button and the ScrollArea behind it.
                     WidgetHits {
-                        contains_pointer: hits,
                         click: Some(hit_click),
                         drag: Some(hit_drag),
+                        ..Default::default()
                     }
                 }
             } else {
-                if hit_drag.sense.click {
+                if hit_drag.sense.senses_click() {
                     // The top thing senses both clicks and drags.
                     WidgetHits {
-                        contains_pointer: hits,
                         click: Some(hit_drag),
                         drag: Some(hit_drag),
+                        ..Default::default()
                     }
                 } else {
                     // The top things senses only drags,
                     // so we ignore the click-widget, because it would be confusing
                     // if clicking a drag-widget would actually click something else below it.
                     WidgetHits {
-                        contains_pointer: hits,
                         click: None,
                         drag: Some(hit_drag),
+                        ..Default::default()
                     }
                 }
             }
@@ -312,8 +408,16 @@ fn hit_test_on_close(close: &[WidgetRect], pos: Pos2) -> WidgetHits {
 }
 
 fn find_closest(widgets: impl Iterator<Item = WidgetRect>, pos: Pos2) -> Option<WidgetRect> {
-    let mut closest = None;
-    let mut closest_dist_sq = f32::INFINITY;
+    find_closest_within(widgets, pos, f32::INFINITY)
+}
+
+fn find_closest_within(
+    widgets: impl Iterator<Item = WidgetRect>,
+    pos: Pos2,
+    max_dist: f32,
+) -> Option<WidgetRect> {
+    let mut closest: Option<WidgetRect> = None;
+    let mut closest_dist_sq = max_dist * max_dist;
     for widget in widgets {
         if widget.interact_rect.is_negative() {
             continue;
@@ -321,6 +425,16 @@ fn find_closest(widgets: impl Iterator<Item = WidgetRect>, pos: Pos2) -> Option<
 
         let dist_sq = widget.interact_rect.distance_sq_to_pos(pos);
 
+        if let Some(closest) = closest {
+            if dist_sq == closest_dist_sq {
+                // It's a tie! Pick the thin candidate over the thick one.
+                // This makes it easier to hit a thin resize-handle, for instance:
+                if should_prioritize_hits_on_back(closest.interact_rect, widget.interact_rect) {
+                    continue;
+                }
+            }
+        }
+
         // In case of a tie, take the last one = the one on top.
         if dist_sq <= closest_dist_sq {
             closest_dist_sq = dist_sq;
@@ -331,9 +445,30 @@ fn find_closest(widgets: impl Iterator<Item = WidgetRect>, pos: Pos2) -> Option<
     closest
 }
 
+/// Should we prioritize hits on `back` over those on `front`?
+///
+/// `back` should be behind the `front` widget.
+///
+/// Returns true if `back` is a small hit-target and `front` is not.
+fn should_prioritize_hits_on_back(back: Rect, front: Rect) -> bool {
+    if front.contains_rect(back) {
+        return false; // back widget is fully occluded; no way to hit it
+    }
+
+    // Reduce each rect to its width or height, whichever is smaller:
+    let back = back.width().min(back.height());
+    let front = front.width().min(front.height());
+
+    // These are hard-coded heuristics that could surely be improved.
+    let back_is_much_thinner = back <= 0.5 * front;
+    let back_is_thin = back <= 16.0;
+
+    back_is_much_thinner && back_is_thin
+}
+
 #[cfg(test)]
 mod tests {
-    use emath::{pos2, vec2, Rect};
+    use emath::{Rect, pos2, vec2};
 
     use crate::{Id, Sense};
 
@@ -385,7 +520,7 @@ mod tests {
         assert_eq!(hits.click.unwrap().id, Id::new("click-and-drag"));
         assert_eq!(hits.drag.unwrap().id, Id::new("click-and-drag"));
 
-        // Close hit - should still ignore the drag-background so as not to confuse the userr:
+        // Close hit - should still ignore the drag-background so as not to confuse the user:
         let hits = hit_test_on_close(&widgets, pos2(105.0, 5.0));
         assert_eq!(hits.click.unwrap().id, Id::new("click-and-drag"));
         assert_eq!(hits.drag.unwrap().id, Id::new("click-and-drag"));
