@@ -1,11 +1,11 @@
 mod touch_state;
 
 use crate::data::input::{
-    Event, EventFilter, KeyboardShortcut, Modifiers, MouseWheelUnit, PointerButton, RawInput,
-    TouchDeviceId, ViewportInfo, NUM_POINTER_BUTTONS,
+    Event, EventFilter, KeyboardShortcut, Modifiers, MouseWheelUnit, NUM_POINTER_BUTTONS,
+    PointerButton, RawInput, TouchDeviceId, ViewportInfo,
 };
 use crate::{
-    emath::{vec2, NumExt, Pos2, Rect, Vec2},
+    emath::{NumExt as _, Pos2, Rect, Vec2, vec2},
     util::History,
 };
 use std::{
@@ -18,9 +18,15 @@ pub use touch_state::MultiTouchInfo;
 use touch_state::TouchState;
 
 /// Options for input state handling.
-#[derive(Clone, Debug, PartialEq)]
+#[derive(Clone, Copy, Debug, PartialEq)]
 #[cfg_attr(feature = "serde", derive(serde::Deserialize, serde::Serialize))]
 pub struct InputOptions {
+    /// Multiplier for the scroll speed when reported in [`crate::MouseWheelUnit::Line`]s.
+    pub line_scroll_speed: f32,
+
+    /// Controls the speed at which we zoom in when doing ctrl/cmd + scroll.
+    pub scroll_zoom_speed: f32,
+
     /// After a pointer-down event, if the pointer moves more than this, it won't become a click.
     pub max_click_dist: f32,
 
@@ -35,14 +41,44 @@ pub struct InputOptions {
     /// The new pointer press must come within this many seconds from previous pointer release
     /// for double click (or when this value is doubled, triple click) to count.
     pub max_double_click_delay: f64,
+
+    /// When this modifier is down, all scroll events are treated as zoom events.
+    ///
+    /// The default is CTRL/CMD, and it is STRONGLY recommended to NOT change this.
+    pub zoom_modifier: Modifiers,
+
+    /// When this modifier is down, all scroll events are treated as horizontal scrolls,
+    /// and when combined with [`Self::zoom_modifier`] it will result in zooming
+    /// on only the horizontal axis.
+    ///
+    /// The default is SHIFT, and it is STRONGLY recommended to NOT change this.
+    pub horizontal_scroll_modifier: Modifiers,
+
+    /// When this modifier is down, all scroll events are treated as vertical scrolls,
+    /// and when combined with [`Self::zoom_modifier`] it will result in zooming
+    /// on only the vertical axis.
+    pub vertical_scroll_modifier: Modifiers,
 }
 
 impl Default for InputOptions {
     fn default() -> Self {
+        // TODO(emilk): figure out why these constants need to be different on web and on native (winit).
+        let is_web = cfg!(target_arch = "wasm32");
+        let line_scroll_speed = if is_web {
+            8.0
+        } else {
+            40.0 // Scroll speed decided by consensus: https://github.com/emilk/egui/issues/461
+        };
+
         Self {
+            line_scroll_speed,
+            scroll_zoom_speed: 1.0 / 200.0,
             max_click_dist: 6.0,
             max_click_duration: 0.8,
             max_double_click_delay: 0.3,
+            zoom_modifier: Modifiers::COMMAND,
+            horizontal_scroll_modifier: Modifiers::SHIFT,
+            vertical_scroll_modifier: Modifiers::ALT,
         }
     }
 }
@@ -51,39 +87,74 @@ impl InputOptions {
     /// Show the options in the ui.
     pub fn ui(&mut self, ui: &mut crate::Ui) {
         let Self {
+            line_scroll_speed,
+            scroll_zoom_speed,
             max_click_dist,
             max_click_duration,
             max_double_click_delay,
+            zoom_modifier,
+            horizontal_scroll_modifier,
+            vertical_scroll_modifier,
         } = self;
-        crate::containers::CollapsingHeader::new("InputOptions")
-            .default_open(false)
+        crate::Grid::new("InputOptions")
+            .num_columns(2)
+            .striped(true)
             .show(ui, |ui| {
-                ui.horizontal(|ui| {
-                    ui.label("Max click distance");
-                    ui.add(
-                        crate::DragValue::new(max_click_dist)
-                            .range(0.0..=f32::INFINITY)
+                ui.label("Line scroll speed");
+                ui.add(crate::DragValue::new(line_scroll_speed).range(0.0..=f32::INFINITY))
+                    .on_hover_text(
+                        "How many lines to scroll with each tick of the mouse wheel",
+                    );
+                ui.end_row();
+
+                ui.label("Scroll zoom speed");
+                ui.add(
+                    crate::DragValue::new(scroll_zoom_speed)
+                        .range(0.0..=f32::INFINITY)
+                        .speed(0.001),
+                )
+                .on_hover_text("How fast to zoom with ctrl/cmd + scroll");
+                ui.end_row();
+
+                ui.label("Max click distance");
+                ui.add(crate::DragValue::new(max_click_dist).range(0.0..=f32::INFINITY))
+                    .on_hover_text(
+                        "If the pointer moves more than this, it won't become a click",
+                    );
+                ui.end_row();
+
+                ui.label("Max click duration");
+                ui.add(
+                    crate::DragValue::new(max_click_duration)
+                        .range(0.1..=f64::INFINITY)
+                        .speed(0.1),
                     )
-                    .on_hover_text("If the pointer moves more than this, it won't become a click");
-                });
-                ui.horizontal(|ui| {
-                    ui.label("Max click duration");
-                    ui.add(
-                        crate::DragValue::new(max_click_duration)
-                            .range(0.1..=f64::INFINITY)
-                            .speed(0.1),
-                    )
-                    .on_hover_text("If the pointer is down for longer than this it will no longer register as a click");
-                });
-                ui.horizontal(|ui| {
-                    ui.label("Max double click delay");
-                    ui.add(
-                        crate::DragValue::new(max_double_click_delay)
-                            .range(0.01..=f64::INFINITY)
-                            .speed(0.1),
-                    )
-                    .on_hover_text("Max time interval for double click to count");
-                });
+                    .on_hover_text(
+                        "If the pointer is down for longer than this it will no longer register as a click",
+                    );
+                ui.end_row();
+
+                ui.label("Max double click delay");
+                ui.add(
+                    crate::DragValue::new(max_double_click_delay)
+                        .range(0.01..=f64::INFINITY)
+                        .speed(0.1),
+                )
+                .on_hover_text("Max time interval for double click to count");
+                ui.end_row();
+
+                ui.label("zoom_modifier");
+                zoom_modifier.ui(ui);
+                ui.end_row();
+
+                ui.label("horizontal_scroll_modifier");
+                horizontal_scroll_modifier.ui(ui);
+                ui.end_row();
+
+                ui.label("vertical_scroll_modifier");
+                vertical_scroll_modifier.ui(ui);
+                ui.end_row();
+
             });
     }
 }
@@ -227,7 +298,7 @@ pub struct InputState {
     /// Input state management configuration.
     ///
     /// This gets copied from `egui::Options` at the start of each frame for convenience.
-    input_options: InputOptions,
+    options: InputOptions,
 }
 
 impl Default for InputState {
@@ -255,7 +326,7 @@ impl Default for InputState {
             modifiers: Default::default(),
             keys_down: Default::default(),
             events: Default::default(),
-            input_options: Default::default(),
+            options: Default::default(),
         }
     }
 }
@@ -267,9 +338,9 @@ impl InputState {
         mut new: RawInput,
         requested_immediate_repaint_prev_frame: bool,
         pixels_per_point: f32,
-        options: &crate::Options,
+        options: InputOptions,
     ) -> Self {
-        crate::profile_function!();
+        profiling::function_scope!();
 
         let time = new.time.unwrap_or(self.time + new.predicted_dt as f64);
         let unstable_dt = (time - self.time) as f32;
@@ -324,11 +395,18 @@ impl InputState {
                         MouseWheelUnit::Page => screen_rect.height() * *delta,
                     };
 
-                    if modifiers.shift {
-                        // Treat as horizontal scrolling.
+                    let is_horizontal = modifiers.matches_any(options.horizontal_scroll_modifier);
+                    let is_vertical = modifiers.matches_any(options.vertical_scroll_modifier);
+
+                    if is_horizontal && !is_vertical {
+                        // Treat all scrolling as horizontal scrolling.
                         // Note: one Mac we already get horizontal scroll events when shift is down.
                         delta = vec2(delta.x + delta.y, 0.0);
                     }
+                    if !is_horizontal && is_vertical {
+                        // Treat all scrolling as vertical scrolling.
+                        delta = vec2(0.0, delta.x + delta.y);
+                    }
 
                     raw_scroll_delta += delta;
 
@@ -342,14 +420,14 @@ impl InputState {
                         MouseWheelUnit::Line | MouseWheelUnit::Page => false,
                     };
 
-                    let is_zoom = modifiers.ctrl || modifiers.mac_cmd || modifiers.command;
+                    let is_zoom = modifiers.matches_any(options.zoom_modifier);
 
-                    #[allow(clippy::collapsible_else_if)]
+                    #[expect(clippy::collapsible_else_if)]
                     if is_zoom {
                         if is_smooth {
-                            smooth_scroll_delta_for_zoom += delta.y;
+                            smooth_scroll_delta_for_zoom += delta.x + delta.y;
                         } else {
-                            unprocessed_scroll_delta_for_zoom += delta.y;
+                            unprocessed_scroll_delta_for_zoom += delta.x + delta.y;
                         }
                     } else {
                         if is_smooth {
@@ -362,6 +440,14 @@ impl InputState {
                 Event::Zoom(factor) => {
                     zoom_factor_delta *= *factor;
                 }
+                Event::WindowFocused(false) => {
+                    // Example: pressing `Cmd+S` brings up a save-dialog (e.g. using rfd),
+                    // but we get no key-up event for the `S` key (in winit).
+                    // This leads to `S` being mistakenly marked as down when we switch back to the app.
+                    // So we take the safe route and just clear all the keys and modifiers when
+                    // the app loses focus.
+                    keys_down.clear();
+                }
                 _ => {}
             }
         }
@@ -429,7 +515,7 @@ impl InputState {
             keys_down,
             events: new.events.clone(), // TODO(emilk): remove clone() and use raw.events
             raw: new,
-            input_options: options.input_options.clone(),
+            options,
         }
     }
 
@@ -444,10 +530,13 @@ impl InputState {
         self.screen_rect
     }
 
-    /// Zoom scale factor this frame (e.g. from ctrl-scroll or pinch gesture).
+    /// Uniform zoom scale factor this frame (e.g. from ctrl-scroll or pinch gesture).
     /// * `zoom = 1`: no change
     /// * `zoom < 1`: pinch together
     /// * `zoom > 1`: pinch spread
+    ///
+    /// If your application supports non-proportional zooming,
+    /// then you probably want to use [`Self::zoom_delta_2d`] instead.
     #[inline(always)]
     pub fn zoom_delta(&self) -> f32 {
         // If a multi touch gesture is detected, it measures the exact and linear proportions of
@@ -477,10 +566,29 @@ impl InputState {
         // the distances of the finger tips.  It is therefore potentially more accurate than
         // `zoom_factor_delta` which is based on the `ctrl-scroll` event which, in turn, may be
         // synthesized from an original touch gesture.
-        self.multi_touch().map_or_else(
-            || Vec2::splat(self.zoom_factor_delta),
-            |touch| touch.zoom_delta_2d,
-        )
+        if let Some(multi_touch) = self.multi_touch() {
+            multi_touch.zoom_delta_2d
+        } else {
+            let mut zoom = Vec2::splat(self.zoom_factor_delta);
+
+            let is_horizontal = self
+                .modifiers
+                .matches_any(self.options.horizontal_scroll_modifier);
+            let is_vertical = self
+                .modifiers
+                .matches_any(self.options.vertical_scroll_modifier);
+
+            if is_horizontal && !is_vertical {
+                // Horizontal-only zooming.
+                zoom.y = 1.0;
+            }
+            if !is_horizontal && is_vertical {
+                // Vertical-only zooming.
+                zoom.x = 1.0;
+            }
+
+            zoom
+        }
     }
 
     /// How long has it been (in seconds) since the use last scrolled?
@@ -489,10 +597,14 @@ impl InputState {
         (self.time - self.last_scroll_time) as f32
     }
 
-    /// The [`crate::Context`] will call this at the end of each frame to see if we need a repaint.
+    /// The [`crate::Context`] will call this at the beginning of each frame to see if we need a repaint.
     ///
     /// Returns how long to wait for a repaint.
-    pub fn wants_repaint_after(&self) -> Option<Duration> {
+    ///
+    /// NOTE: It's important to call this immediately after [`Self::begin_pass`] since calls to
+    /// [`Self::consume_key`] will remove events from the vec, meaning those key presses wouldn't
+    /// cause a repaint.
+    pub(crate) fn wants_repaint_after(&self) -> Option<Duration> {
         if self.pointer.wants_repaint()
             || self.unprocessed_scroll_delta.abs().max_elem() > 0.2
             || self.unprocessed_scroll_delta_for_zoom.abs() > 0.2
@@ -506,10 +618,10 @@ impl InputState {
             // We need to wake up and check for press-and-hold for the context menu.
             if let Some(press_start_time) = self.pointer.press_start_time {
                 let press_duration = self.time - press_start_time;
-                if self.input_options.max_click_duration.is_finite()
-                    && press_duration < self.input_options.max_click_duration
+                if self.options.max_click_duration.is_finite()
+                    && press_duration < self.options.max_click_duration
                 {
-                    let secs_until_menu = self.input_options.max_click_duration - press_duration;
+                    let secs_until_menu = self.options.max_click_duration - press_duration;
                     return Some(Duration::from_secs_f64(secs_until_menu));
                 }
             }
@@ -712,6 +824,23 @@ impl InputState {
         })
     }
 
+    #[cfg(feature = "accesskit")]
+    pub fn consume_accesskit_action_requests(
+        &mut self,
+        id: crate::Id,
+        mut consume: impl FnMut(&accesskit::ActionRequest) -> bool,
+    ) {
+        let accesskit_id = id.accesskit_id();
+        self.events.retain(|event| {
+            if let Event::AccessKitActionRequest(request) = event {
+                if request.target == accesskit_id {
+                    return !consume(request);
+                }
+            }
+            true
+        });
+    }
+
     #[cfg(feature = "accesskit")]
     pub fn has_accesskit_action_request(&self, id: crate::Id, action: accesskit::Action) -> bool {
         self.accesskit_action_requests(id, action).next().is_some()
@@ -870,7 +999,7 @@ pub struct PointerState {
     /// Input state management configuration.
     ///
     /// This gets copied from `egui::Options` at the start of each frame for convenience.
-    input_options: InputOptions,
+    options: InputOptions,
 }
 
 impl Default for PointerState {
@@ -889,27 +1018,22 @@ impl Default for PointerState {
             press_start_time: None,
             has_moved_too_much_for_a_click: false,
             started_decidedly_dragging: false,
-            last_click_time: std::f64::NEG_INFINITY,
-            last_last_click_time: std::f64::NEG_INFINITY,
-            last_move_time: std::f64::NEG_INFINITY,
+            last_click_time: f64::NEG_INFINITY,
+            last_last_click_time: f64::NEG_INFINITY,
+            last_move_time: f64::NEG_INFINITY,
             pointer_events: vec![],
-            input_options: Default::default(),
+            options: Default::default(),
         }
     }
 }
 
 impl PointerState {
     #[must_use]
-    pub(crate) fn begin_pass(
-        mut self,
-        time: f64,
-        new: &RawInput,
-        options: &crate::Options,
-    ) -> Self {
+    pub(crate) fn begin_pass(mut self, time: f64, new: &RawInput, options: InputOptions) -> Self {
         let was_decidedly_dragging = self.is_decidedly_dragging();
 
         self.time = time;
-        self.input_options = options.input_options.clone();
+        self.options = options;
 
         self.pointer_events.clear();
 
@@ -919,6 +1043,7 @@ impl PointerState {
             self.motion = Some(Vec2::ZERO);
         }
 
+        let mut clear_history_after_velocity_calculation = false;
         for event in &new.events {
             match event {
                 Event::PointerMoved(pos) => {
@@ -929,9 +1054,10 @@ impl PointerState {
 
                     if let Some(press_origin) = self.press_origin {
                         self.has_moved_too_much_for_a_click |=
-                            press_origin.distance(pos) > self.input_options.max_click_dist;
+                            press_origin.distance(pos) > self.options.max_click_dist;
                     }
 
+                    self.last_move_time = time;
                     self.pointer_events.push(PointerEvent::Moved(pos));
                 }
                 Event::PointerButton {
@@ -967,10 +1093,10 @@ impl PointerState {
                         let clicked = self.could_any_button_be_click();
 
                         let click = if clicked {
-                            let double_click = (time - self.last_click_time)
-                                < self.input_options.max_double_click_delay;
+                            let double_click =
+                                (time - self.last_click_time) < self.options.max_double_click_delay;
                             let triple_click = (time - self.last_last_click_time)
-                                < (self.input_options.max_double_click_delay * 2.0);
+                                < (self.options.max_double_click_delay * 2.0);
                             let count = if triple_click {
                                 3
                             } else if double_click {
@@ -1005,7 +1131,10 @@ impl PointerState {
                     // When dragging a slider and the mouse leaves the viewport, we still want the drag to work,
                     // so we don't treat this as a `PointerEvent::Released`.
                     // NOTE: we do NOT clear `self.interact_pos` here. It will be cleared next frame.
-                    self.pos_history.clear();
+
+                    // Delay the clearing until after the final velocity calculation, so we can
+                    // get the final velocity when `drag_stopped` is true.
+                    clear_history_after_velocity_calculation = true;
                 }
                 Event::MouseMoved(delta) => *self.motion.get_or_insert(Vec2::ZERO) += *delta,
                 _ => {}
@@ -1036,6 +1165,9 @@ impl PointerState {
         if self.velocity != Vec2::ZERO {
             self.last_move_time = time;
         }
+        if clear_history_after_velocity_calculation {
+            self.pos_history.clear();
+        }
 
         self.direction = self.pos_history.velocity().unwrap_or_default().normalized();
 
@@ -1268,7 +1400,7 @@ impl PointerState {
             }
 
             if let Some(press_start_time) = self.press_start_time {
-                if self.time - press_start_time > self.input_options.max_click_duration {
+                if self.time - press_start_time > self.options.max_click_duration {
                     return false;
                 }
             }
@@ -1303,8 +1435,8 @@ impl PointerState {
         self.started_decidedly_dragging
             && !self.has_moved_too_much_for_a_click
             && self.button_down(PointerButton::Primary)
-            && self.press_start_time.map_or(false, |press_start_time| {
-                self.time - press_start_time > self.input_options.max_click_duration
+            && self.press_start_time.is_some_and(|press_start_time| {
+                self.time - press_start_time > self.options.max_click_duration
             })
     }
 
@@ -1325,6 +1457,21 @@ impl PointerState {
     pub fn middle_down(&self) -> bool {
         self.button_down(PointerButton::Middle)
     }
+
+    /// Is the mouse moving in the direction of the given rect?
+    pub fn is_moving_towards_rect(&self, rect: &Rect) -> bool {
+        if self.is_still() {
+            return false;
+        }
+
+        if let Some(pos) = self.hover_pos() {
+            let dir = self.direction();
+            if dir != Vec2::ZERO {
+                return rect.intersects_ray(pos, self.direction());
+            }
+        }
+        false
+    }
 }
 
 impl InputState {
@@ -1352,7 +1499,7 @@ impl InputState {
             modifiers,
             keys_down,
             events,
-            input_options: _,
+            options: _,
         } = self;
 
         ui.style_mut()
@@ -1438,7 +1585,7 @@ impl PointerState {
             last_last_click_time,
             pointer_events,
             last_move_time,
-            input_options: _,
+            options: _,
         } = self;
 
         ui.label(format!("latest_pos: {latest_pos:?}"));
