@@ -1,22 +1,23 @@
 use std::{any::Any, sync::Arc};
 
 use crate::{
-    emath::{Align, Pos2, Rect, Vec2},
-    menu, pass_state, AreaState, Context, CursorIcon, Id, LayerId, Order, PointerButton, Sense, Ui,
+    Context, CursorIcon, Id, LayerId, PointerButton, Popup, PopupKind, Sense, Tooltip, Ui,
     WidgetRect, WidgetText,
+    emath::{Align, Pos2, Rect, Vec2},
+    pass_state,
 };
 // ----------------------------------------------------------------------------
 
 /// The result of adding a widget to a [`Ui`].
 ///
-/// A [`Response`] lets you know whether or not a widget is being hovered, clicked or dragged.
+/// A [`Response`] lets you know whether a widget is being hovered, clicked or dragged.
 /// It also lets you easily show a tooltip on hover.
 ///
 /// Whenever something gets added to a [`Ui`], a [`Response`] object is returned.
 /// [`ui.add`] returns a [`Response`], as does [`ui.button`], and all similar shortcuts.
 ///
 /// ⚠️ The `Response` contains a clone of [`Context`], and many methods lock the `Context`.
-/// It can therefor be a deadlock to use `Context` from within a context-locking closures,
+/// It can therefore be a deadlock to use `Context` from within a context-locking closures,
 /// such as [`Context::input`].
 #[derive(Clone, Debug)]
 pub struct Response {
@@ -50,82 +51,16 @@ pub struct Response {
     /// (that is handled by the `Painter` directly).
     pub sense: Sense,
 
-    /// Was the widget enabled?
-    /// If `false`, there was no interaction attempted (not even hover).
-    #[doc(hidden)]
-    pub enabled: bool,
-
     // OUT:
-    /// The pointer is above this widget with no other blocking it.
-    #[doc(hidden)]
-    pub contains_pointer: bool,
-
-    /// The pointer is hovering above this widget or the widget was clicked/tapped this frame.
-    #[doc(hidden)]
-    pub hovered: bool,
-
-    /// The widget is highlighted via a call to [`Self::highlight`] or [`Context::highlight_widget`].
-    #[doc(hidden)]
-    pub highlighted: bool,
-
-    /// This widget was clicked this frame.
-    ///
-    /// Which pointer and how many times we don't know,
-    /// and ask [`crate::InputState`] about at runtime.
-    ///
-    /// This is only set to true if the widget was clicked
-    /// by an actual mouse.
-    #[doc(hidden)]
-    pub clicked: bool,
-
-    /// This widget should act as if clicked due
-    /// to something else than a click.
-    ///
-    /// This is set to true if the widget has keyboard focus and
-    /// the user hit the Space or Enter key.
-    #[doc(hidden)]
-    pub fake_primary_click: bool,
-
-    /// This widget was long-pressed on a touch screen to simulate a secondary click.
-    #[doc(hidden)]
-    pub long_touched: bool,
-
-    /// The widget started being dragged this frame.
-    #[doc(hidden)]
-    pub drag_started: bool,
-
-    /// The widget is being dragged.
-    #[doc(hidden)]
-    pub dragged: bool,
-
-    /// The widget was being dragged, but now it has been released.
-    #[doc(hidden)]
-    pub drag_stopped: bool,
-
-    /// Is the pointer button currently down on this widget?
-    /// This is true if the pointer is pressing down or dragging a widget
-    #[doc(hidden)]
-    pub is_pointer_button_down_on: bool,
-
-    /// Where the pointer (mouse/touch) were when when this widget was clicked or dragged.
+    /// Where the pointer (mouse/touch) were when this widget was clicked or dragged.
     /// `None` if the widget is not being interacted with.
     #[doc(hidden)]
     pub interact_pointer_pos: Option<Pos2>,
 
-    /// Was the underlying data changed?
-    ///
-    /// e.g. the slider was dragged, text was entered in a [`TextEdit`](crate::TextEdit) etc.
-    /// Always `false` for something like a [`Button`](crate::Button).
-    ///
-    /// Note that this can be `true` even if the user did not interact with the widget,
-    /// for instance if an existing slider value was clamped to the given range.
-    #[doc(hidden)]
-    pub changed: bool,
-
     /// The intrinsic / desired size of the widget.
     ///
-    /// For a button, this will be the size of the label + the frames padding,
-    /// even if the button is laid out in a justified layout and the actual size will be larger.
+    /// This is the size that a non-wrapped, non-truncated, non-justified version of the widget
+    /// would have.
     ///
     /// If this is `None`, use [`Self::rect`] instead.
     ///
@@ -133,6 +68,76 @@ pub struct Response {
     /// for improved layouting.
     /// See for instance [`egui_flex`](https://github.com/lucasmerlin/hello_egui/tree/main/crates/egui_flex).
     pub intrinsic_size: Option<Vec2>,
+
+    #[doc(hidden)]
+    pub flags: Flags,
+}
+
+/// A bit set for various boolean properties of `Response`.
+#[doc(hidden)]
+#[derive(Copy, Clone, Debug)]
+pub struct Flags(u16);
+
+bitflags::bitflags! {
+    impl Flags: u16 {
+        /// Was the widget enabled?
+        /// If `false`, there was no interaction attempted (not even hover).
+        const ENABLED = 1<<0;
+
+        /// The pointer is above this widget with no other blocking it.
+        const CONTAINS_POINTER = 1<<1;
+
+        /// The pointer is hovering above this widget or the widget was clicked/tapped this frame.
+        const HOVERED = 1<<2;
+
+        /// The widget is highlighted via a call to [`Response::highlight`] or
+        /// [`Context::highlight_widget`].
+        const HIGHLIGHTED = 1<<3;
+
+        /// This widget was clicked this frame.
+        ///
+        /// Which pointer and how many times we don't know,
+        /// and ask [`crate::InputState`] about at runtime.
+        ///
+        /// This is only set to true if the widget was clicked
+        /// by an actual mouse.
+        const CLICKED = 1<<4;
+
+        /// This widget should act as if clicked due
+        /// to something else than a click.
+        ///
+        /// This is set to true if the widget has keyboard focus and
+        /// the user hit the Space or Enter key.
+        const FAKE_PRIMARY_CLICKED = 1<<5;
+
+        /// This widget was long-pressed on a touch screen to simulate a secondary click.
+        const LONG_TOUCHED = 1<<6;
+
+        /// The widget started being dragged this frame.
+        const DRAG_STARTED = 1<<7;
+
+        /// The widget is being dragged.
+        const DRAGGED = 1<<8;
+
+        /// The widget was being dragged, but now it has been released.
+        const DRAG_STOPPED = 1<<9;
+
+        /// Is the pointer button currently down on this widget?
+        /// This is true if the pointer is pressing down or dragging a widget
+        const IS_POINTER_BUTTON_DOWN_ON = 1<<10;
+
+        /// Was the underlying data changed?
+        ///
+        /// e.g. the slider was dragged, text was entered in a [`TextEdit`](crate::TextEdit) etc.
+        /// Always `false` for something like a [`Button`](crate::Button).
+        ///
+        /// Note that this can be `true` even if the user did not interact with the widget,
+        /// for instance if an existing slider value was clamped to the given range.
+        const CHANGED = 1<<11;
+
+        /// Should this container be closed?
+        const CLOSE = 1<<12;
+    }
 }
 
 impl Response {
@@ -150,7 +155,7 @@ impl Response {
     /// You can use [`Self::interact`] to sense more things *after* adding a widget.
     #[inline(always)]
     pub fn clicked(&self) -> bool {
-        self.fake_primary_click || self.clicked_by(PointerButton::Primary)
+        self.flags.contains(Flags::FAKE_PRIMARY_CLICKED) || self.clicked_by(PointerButton::Primary)
     }
 
     /// Returns true if this widget was clicked this frame by the given mouse button.
@@ -163,7 +168,7 @@ impl Response {
     /// Use [`Self::secondary_clicked`] instead to also detect that.
     #[inline]
     pub fn clicked_by(&self, button: PointerButton) -> bool {
-        self.clicked && self.ctx.input(|i| i.pointer.button_clicked(button))
+        self.flags.contains(Flags::CLICKED) && self.ctx.input(|i| i.pointer.button_clicked(button))
     }
 
     /// Returns true if this widget was clicked this frame by the secondary mouse button (e.g. the right mouse button).
@@ -171,7 +176,7 @@ impl Response {
     /// This also returns true if the widget was pressed-and-held on a touch screen.
     #[inline]
     pub fn secondary_clicked(&self) -> bool {
-        self.long_touched || self.clicked_by(PointerButton::Secondary)
+        self.flags.contains(Flags::LONG_TOUCHED) || self.clicked_by(PointerButton::Secondary)
     }
 
     /// Was this long-pressed on a touch screen?
@@ -179,7 +184,7 @@ impl Response {
     /// Usually you want to check [`Self::secondary_clicked`] instead.
     #[inline]
     pub fn long_touched(&self) -> bool {
-        self.long_touched
+        self.flags.contains(Flags::LONG_TOUCHED)
     }
 
     /// Returns true if this widget was clicked this frame by the middle mouse button.
@@ -203,38 +208,55 @@ impl Response {
     /// Returns true if this widget was double-clicked this frame by the given button.
     #[inline]
     pub fn double_clicked_by(&self, button: PointerButton) -> bool {
-        self.clicked && self.ctx.input(|i| i.pointer.button_double_clicked(button))
+        self.flags.contains(Flags::CLICKED)
+            && self.ctx.input(|i| i.pointer.button_double_clicked(button))
     }
 
     /// Returns true if this widget was triple-clicked this frame by the given button.
     #[inline]
     pub fn triple_clicked_by(&self, button: PointerButton) -> bool {
-        self.clicked && self.ctx.input(|i| i.pointer.button_triple_clicked(button))
+        self.flags.contains(Flags::CLICKED)
+            && self.ctx.input(|i| i.pointer.button_triple_clicked(button))
+    }
+
+    /// Was this widget middle-clicked or clicked while holding down a modifier key?
+    ///
+    /// This is used by [`crate::Hyperlink`] to check if a URL should be opened
+    /// in a new tab, using [`crate::OpenUrl::new_tab`].
+    pub fn clicked_with_open_in_background(&self) -> bool {
+        self.middle_clicked() || self.clicked() && self.ctx.input(|i| i.modifiers.any())
     }
 
     /// `true` if there was a click *outside* the rect of this widget.
     ///
     /// Clicks on widgets contained in this one counts as clicks inside this widget,
     /// so that clicking a button in an area will not be considered as clicking "elsewhere" from the area.
+    ///
+    /// Clicks on other layers above this widget *will* be considered as clicking elsewhere.
     pub fn clicked_elsewhere(&self) -> bool {
+        let (pointer_interact_pos, any_click) = self
+            .ctx
+            .input(|i| (i.pointer.interact_pos(), i.pointer.any_click()));
+
         // We do not use self.clicked(), because we want to catch all clicks within our frame,
         // even if we aren't clickable (or even enabled).
         // This is important for windows and such that should close then the user clicks elsewhere.
-        self.ctx.input(|i| {
-            let pointer = &i.pointer;
-
-            if pointer.any_click() {
-                if self.contains_pointer || self.hovered {
-                    false
-                } else if let Some(pos) = pointer.interact_pos() {
-                    !self.interact_rect.contains(pos)
+        if any_click {
+            if self.contains_pointer() || self.hovered() {
+                false
+            } else if let Some(pos) = pointer_interact_pos {
+                let layer_under_pointer = self.ctx.layer_id_at(pos);
+                if layer_under_pointer != Some(self.layer_id) {
+                    true
                 } else {
-                    false // clicked without a pointer, weird
+                    !self.interact_rect.contains(pos)
                 }
             } else {
-                false
+                false // clicked without a pointer, weird
             }
-        })
+        } else {
+            false
+        }
     }
 
     /// Was the widget enabled?
@@ -242,7 +264,7 @@ impl Response {
     /// and the widget should be drawn in a gray disabled look.
     #[inline(always)]
     pub fn enabled(&self) -> bool {
-        self.enabled
+        self.flags.contains(Flags::ENABLED)
     }
 
     /// The pointer is hovering above this widget or the widget was clicked/tapped this frame.
@@ -251,7 +273,7 @@ impl Response {
     /// `hovered` is always `false` for disabled widgets.
     #[inline(always)]
     pub fn hovered(&self) -> bool {
-        self.hovered
+        self.flags.contains(Flags::HOVERED)
     }
 
     /// Returns true if the pointer is contained by the response rect, and no other widget is covering it.
@@ -264,14 +286,14 @@ impl Response {
     /// [`Self::contains_pointer`] also checks that no other widget is covering this response rectangle.
     #[inline(always)]
     pub fn contains_pointer(&self) -> bool {
-        self.contains_pointer
+        self.flags.contains(Flags::CONTAINS_POINTER)
     }
 
     /// The widget is highlighted via a call to [`Self::highlight`] or [`Context::highlight_widget`].
     #[doc(hidden)]
     #[inline(always)]
     pub fn highlighted(&self) -> bool {
-        self.highlighted
+        self.flags.contains(Flags::HIGHLIGHTED)
     }
 
     /// This widget has the keyboard focus (i.e. is receiving key presses).
@@ -316,7 +338,7 @@ impl Response {
         self.ctx.memory_mut(|mem| mem.surrender_focus(self.id));
     }
 
-    /// Did a drag on this widgets begin this frame?
+    /// Did a drag on this widget begin this frame?
     ///
     /// This is only true if the widget sense drags.
     /// If the widget also senses clicks, this will only become true if the pointer has moved a bit.
@@ -324,10 +346,10 @@ impl Response {
     /// This will only be true for a single frame.
     #[inline]
     pub fn drag_started(&self) -> bool {
-        self.drag_started
+        self.flags.contains(Flags::DRAG_STARTED)
     }
 
-    /// Did a drag on this widgets by the button begin this frame?
+    /// Did a drag on this widget by the button begin this frame?
     ///
     /// This is only true if the widget sense drags.
     /// If the widget also senses clicks, this will only become true if the pointer has moved a bit.
@@ -354,7 +376,7 @@ impl Response {
     /// You can use [`Self::interact`] to sense more things *after* adding a widget.
     #[inline(always)]
     pub fn dragged(&self) -> bool {
-        self.dragged
+        self.flags.contains(Flags::DRAGGED)
     }
 
     /// See [`Self::dragged`].
@@ -366,7 +388,7 @@ impl Response {
     /// The widget was being dragged, but now it has been released.
     #[inline]
     pub fn drag_stopped(&self) -> bool {
-        self.drag_stopped
+        self.flags.contains(Flags::DRAG_STOPPED)
     }
 
     /// The widget was being dragged by the button, but now it has been released.
@@ -374,29 +396,13 @@ impl Response {
         self.drag_stopped() && self.ctx.input(|i| i.pointer.button_released(button))
     }
 
-    /// The widget was being dragged, but now it has been released.
-    #[inline]
-    #[deprecated = "Renamed 'drag_stopped'"]
-    pub fn drag_released(&self) -> bool {
-        self.drag_stopped
-    }
-
-    /// The widget was being dragged by the button, but now it has been released.
-    #[deprecated = "Renamed 'drag_stopped_by'"]
-    pub fn drag_released_by(&self, button: PointerButton) -> bool {
-        self.drag_stopped_by(button)
-    }
-
     /// If dragged, how many points were we dragged and in what direction?
     #[inline]
     pub fn drag_delta(&self) -> Vec2 {
         if self.dragged() {
             let mut delta = self.ctx.input(|i| i.pointer.delta());
-            if let Some(scaling) = self
-                .ctx
-                .memory(|m| m.layer_transforms.get(&self.layer_id).map(|t| t.scaling))
-            {
-                delta /= scaling;
+            if let Some(from_global) = self.ctx.layer_transform_from_global(self.layer_id) {
+                delta *= from_global.scaling;
             }
             delta
         } else {
@@ -425,7 +431,7 @@ impl Response {
             crate::DragAndDrop::set_payload(&self.ctx, payload);
         }
 
-        if self.hovered() && !self.sense.click {
+        if self.hovered() && !self.sense.senses_click() {
             // Things that can be drag-dropped should use the Grab cursor icon,
             // but if the thing is _also_ clickable, that can be annoying.
             self.ctx.set_cursor_icon(CursorIcon::Grab);
@@ -463,7 +469,7 @@ impl Response {
         }
     }
 
-    /// Where the pointer (mouse/touch) were when when this widget was clicked or dragged.
+    /// Where the pointer (mouse/touch) were when this widget was clicked or dragged.
     ///
     /// `None` if the widget is not being interacted with.
     #[inline]
@@ -478,11 +484,8 @@ impl Response {
     pub fn hover_pos(&self) -> Option<Pos2> {
         if self.hovered() {
             let mut pos = self.ctx.input(|i| i.pointer.hover_pos())?;
-            if let Some(transform) = self
-                .ctx
-                .memory(|m| m.layer_transforms.get(&self.layer_id).copied())
-            {
-                pos = transform.inverse() * pos;
+            if let Some(from_global) = self.ctx.layer_transform_from_global(self.layer_id) {
+                pos = from_global * pos;
             }
             Some(pos)
         } else {
@@ -498,7 +501,7 @@ impl Response {
     /// This could also be thought of as "is this widget being interacted with?".
     #[inline(always)]
     pub fn is_pointer_button_down_on(&self) -> bool {
-        self.is_pointer_button_down_on
+        self.flags.contains(Flags::IS_POINTER_BUTTON_DOWN_ON)
     }
 
     /// Was the underlying data changed?
@@ -516,7 +519,7 @@ impl Response {
     /// for instance if an existing slider value was clamped to the given range.
     #[inline(always)]
     pub fn changed(&self) -> bool {
-        self.changed
+        self.flags.contains(Flags::CHANGED)
     }
 
     /// Report the data shown by this widget changed.
@@ -525,10 +528,25 @@ impl Response {
     /// e.g. checkboxes, sliders etc.
     ///
     /// This should be called when the *content* changes, but not when the view does.
-    /// So we call this when the text of a [`crate::TextEdit`], but not when the cursors changes.
+    /// So we call this when the text of a [`crate::TextEdit`], but not when the cursor changes.
     #[inline(always)]
     pub fn mark_changed(&mut self) {
-        self.changed = true;
+        self.flags.set(Flags::CHANGED, true);
+    }
+
+    /// Should the container be closed?
+    ///
+    /// Will e.g. be set by calling [`Ui::close`] in a child [`Ui`] or by calling
+    /// [`Self::set_close`].
+    pub fn should_close(&self) -> bool {
+        self.flags.contains(Flags::CLOSE)
+    }
+
+    /// Set the [`Flags::CLOSE`] flag.
+    ///
+    /// Can be used to e.g. signal that a container should be closed.
+    pub fn set_close(&mut self) {
+        self.flags.set(Flags::CLOSE, true);
     }
 
     /// Show this UI if the widget was hovered (i.e. a tooltip).
@@ -553,36 +571,22 @@ impl Response {
     /// ```
     #[doc(alias = "tooltip")]
     pub fn on_hover_ui(self, add_contents: impl FnOnce(&mut Ui)) -> Self {
-        if self.enabled && self.should_show_hover_ui() {
-            self.show_tooltip_ui(add_contents);
-        }
+        Tooltip::for_enabled(&self).show(add_contents);
         self
     }
 
     /// Show this UI when hovering if the widget is disabled.
     pub fn on_disabled_hover_ui(self, add_contents: impl FnOnce(&mut Ui)) -> Self {
-        if !self.enabled && self.should_show_hover_ui() {
-            crate::containers::show_tooltip_for(
-                &self.ctx,
-                self.layer_id,
-                self.id,
-                &self.rect,
-                add_contents,
-            );
-        }
+        Tooltip::for_disabled(&self).show(add_contents);
         self
     }
 
     /// Like `on_hover_ui`, but show the ui next to cursor.
     pub fn on_hover_ui_at_pointer(self, add_contents: impl FnOnce(&mut Ui)) -> Self {
-        if self.enabled && self.should_show_hover_ui() {
-            crate::containers::show_tooltip_at_pointer(
-                &self.ctx,
-                self.layer_id,
-                self.id,
-                add_contents,
-            );
-        }
+        Tooltip::for_enabled(&self)
+            .at_pointer()
+            .gap(12.0)
+            .show(add_contents);
         self
     }
 
@@ -590,13 +594,9 @@ impl Response {
     ///
     /// This can be used to give attention to a widget during a tutorial.
     pub fn show_tooltip_ui(&self, add_contents: impl FnOnce(&mut Ui)) {
-        crate::containers::show_tooltip_for(
-            &self.ctx,
-            self.layer_id,
-            self.id,
-            &self.rect,
-            add_contents,
-        );
+        Popup::from_response(self)
+            .kind(PopupKind::Tooltip)
+            .show(add_contents);
     }
 
     /// Always show this tooltip, even if disabled and the user isn't hovering it.
@@ -610,180 +610,7 @@ impl Response {
 
     /// Was the tooltip open last frame?
     pub fn is_tooltip_open(&self) -> bool {
-        crate::popup::was_tooltip_open_last_frame(&self.ctx, self.id)
-    }
-
-    fn should_show_hover_ui(&self) -> bool {
-        if self.ctx.memory(|mem| mem.everything_is_visible()) {
-            return true;
-        }
-
-        let any_open_popups = self.ctx.prev_pass_state(|fs| {
-            fs.layers
-                .get(&self.layer_id)
-                .map_or(false, |layer| !layer.open_popups.is_empty())
-        });
-        if any_open_popups {
-            // Hide tooltips if the user opens a popup (menu, combo-box, etc) in the same layer.
-            return false;
-        }
-
-        let style = self.ctx.style();
-
-        let tooltip_delay = style.interaction.tooltip_delay;
-        let tooltip_grace_time = style.interaction.tooltip_grace_time;
-
-        let (
-            time_since_last_scroll,
-            time_since_last_click,
-            time_since_last_pointer_movement,
-            pointer_pos,
-            pointer_dir,
-        ) = self.ctx.input(|i| {
-            (
-                i.time_since_last_scroll(),
-                i.pointer.time_since_last_click(),
-                i.pointer.time_since_last_movement(),
-                i.pointer.hover_pos(),
-                i.pointer.direction(),
-            )
-        });
-
-        if time_since_last_scroll < tooltip_delay {
-            // See https://github.com/emilk/egui/issues/4781
-            // Note that this means we cannot have `ScrollArea`s in a tooltip.
-            self.ctx
-                .request_repaint_after_secs(tooltip_delay - time_since_last_scroll);
-            return false;
-        }
-
-        let is_our_tooltip_open = self.is_tooltip_open();
-
-        if is_our_tooltip_open {
-            // Check if we should automatically stay open:
-
-            let tooltip_id = crate::next_tooltip_id(&self.ctx, self.id);
-            let tooltip_layer_id = LayerId::new(Order::Tooltip, tooltip_id);
-
-            let tooltip_has_interactive_widget = self.ctx.viewport(|vp| {
-                vp.prev_pass
-                    .widgets
-                    .get_layer(tooltip_layer_id)
-                    .any(|w| w.enabled && w.sense.interactive())
-            });
-
-            if tooltip_has_interactive_widget {
-                // We keep the tooltip open if hovered,
-                // or if the pointer is on its way to it,
-                // so that the user can interact with the tooltip
-                // (i.e. click links that are in it).
-                if let Some(area) = AreaState::load(&self.ctx, tooltip_id) {
-                    let rect = area.rect();
-
-                    if let Some(pos) = pointer_pos {
-                        if rect.contains(pos) {
-                            return true; // hovering interactive tooltip
-                        }
-                        if pointer_dir != Vec2::ZERO
-                            && rect.intersects_ray(pos, pointer_dir.normalized())
-                        {
-                            return true; // on the way to interactive tooltip
-                        }
-                    }
-                }
-            }
-        }
-
-        let clicked_more_recently_than_moved =
-            time_since_last_click < time_since_last_pointer_movement + 0.1;
-        if clicked_more_recently_than_moved {
-            // It is common to click a widget and then rest the mouse there.
-            // It would be annoying to then see a tooltip for it immediately.
-            // Similarly, clicking should hide the existing tooltip.
-            // Only hovering should lead to a tooltip, not clicking.
-            // The offset is only to allow small movement just right after the click.
-            return false;
-        }
-
-        if is_our_tooltip_open {
-            // Check if we should automatically stay open:
-
-            if pointer_pos.is_some_and(|pointer_pos| self.rect.contains(pointer_pos)) {
-                // Handle the case of a big tooltip that covers the widget:
-                return true;
-            }
-        }
-
-        let is_other_tooltip_open = self.ctx.prev_pass_state(|fs| {
-            if let Some(already_open_tooltip) = fs
-                .layers
-                .get(&self.layer_id)
-                .and_then(|layer| layer.widget_with_tooltip)
-            {
-                already_open_tooltip != self.id
-            } else {
-                false
-            }
-        });
-        if is_other_tooltip_open {
-            // We only allow one tooltip per layer. First one wins. It is up to that tooltip to close itself.
-            return false;
-        }
-
-        // Fast early-outs:
-        if self.enabled {
-            if !self.hovered || !self.ctx.input(|i| i.pointer.has_pointer()) {
-                return false;
-            }
-        } else if !self.ctx.rect_contains_pointer(self.layer_id, self.rect) {
-            return false;
-        }
-
-        // There is a tooltip_delay before showing the first tooltip,
-        // but once one tooltips is show, moving the mouse cursor to
-        // another widget should show the tooltip for that widget right away.
-
-        // Let the user quickly move over some dead space to hover the next thing
-        let tooltip_was_recently_shown =
-            crate::popup::seconds_since_last_tooltip(&self.ctx) < tooltip_grace_time;
-
-        if !tooltip_was_recently_shown && !is_our_tooltip_open {
-            if style.interaction.show_tooltips_only_when_still {
-                // We only show the tooltip when the mouse pointer is still.
-                if !self
-                    .ctx
-                    .input(|i| i.pointer.is_still() && i.smooth_scroll_delta == Vec2::ZERO)
-                {
-                    // wait for mouse to stop
-                    self.ctx.request_repaint();
-                    return false;
-                }
-            }
-
-            let time_since_last_interaction = time_since_last_scroll
-                .min(time_since_last_pointer_movement)
-                .min(time_since_last_click);
-            let time_til_tooltip = tooltip_delay - time_since_last_interaction;
-
-            if 0.0 < time_til_tooltip {
-                // Wait until the mouse has been still for a while
-                self.ctx.request_repaint_after_secs(time_til_tooltip);
-                return false;
-            }
-        }
-
-        // We don't want tooltips of things while we are dragging them,
-        // but we do want tooltips while holding down on an item on a touch screen.
-        if self
-            .ctx
-            .input(|i| i.pointer.any_down() && i.pointer.has_moved_too_much_for_a_click)
-        {
-            return false;
-        }
-
-        // All checks passed: show the tooltip!
-
-        true
+        Tooltip::was_tooltip_open_last_frame(&self.ctx, self.id)
     }
 
     /// Like `on_hover_text`, but show the text next to cursor.
@@ -823,7 +650,7 @@ impl Response {
     #[inline]
     pub fn highlight(mut self) -> Self {
         self.ctx.highlight_widget(self.id);
-        self.highlighted = true;
+        self.flags.set(Flags::HIGHLIGHTED, true);
         self
     }
 
@@ -894,7 +721,7 @@ impl Response {
                 rect: self.rect,
                 interact_rect: self.interact_rect,
                 sense: self.sense | sense,
-                enabled: self.enabled,
+                enabled: self.enabled(),
             },
             true,
         )
@@ -957,7 +784,7 @@ impl Response {
             Some(OutputEvent::TripleClicked(make_info()))
         } else if self.gained_focus() {
             Some(OutputEvent::FocusGained(make_info()))
-        } else if self.changed {
+        } else if self.changed() {
             Some(OutputEvent::ValueChanged(make_info()))
         } else {
             None
@@ -988,8 +815,8 @@ impl Response {
     }
 
     #[cfg(feature = "accesskit")]
-    pub(crate) fn fill_accesskit_node_common(&self, builder: &mut accesskit::NodeBuilder) {
-        if !self.enabled {
+    pub(crate) fn fill_accesskit_node_common(&self, builder: &mut accesskit::Node) {
+        if !self.enabled() {
             builder.set_disabled();
         }
         builder.set_bounds(accesskit::Rect {
@@ -998,18 +825,18 @@ impl Response {
             x1: self.rect.max.x.into(),
             y1: self.rect.max.y.into(),
         });
-        if self.sense.focusable {
+        if self.sense.is_focusable() {
             builder.add_action(accesskit::Action::Focus);
         }
-        if self.sense.click && builder.default_action_verb().is_none() {
-            builder.set_default_action_verb(accesskit::DefaultActionVerb::Click);
+        if self.sense.senses_click() {
+            builder.add_action(accesskit::Action::Click);
         }
     }
 
     #[cfg(feature = "accesskit")]
     fn fill_accesskit_node_from_widget_info(
         &self,
-        builder: &mut accesskit::NodeBuilder,
+        builder: &mut accesskit::Node,
         info: crate::WidgetInfo,
     ) {
         use crate::WidgetType;
@@ -1023,6 +850,7 @@ impl Response {
             WidgetType::Button | WidgetType::ImageButton | WidgetType::CollapsingHeader => {
                 Role::Button
             }
+            WidgetType::Image => Role::Image,
             WidgetType::Checkbox => Role::CheckBox,
             WidgetType::RadioButton => Role::RadioButton,
             WidgetType::RadioGroup => Role::RadioGroup,
@@ -1032,13 +860,18 @@ impl Response {
             WidgetType::DragValue => Role::SpinButton,
             WidgetType::ColorButton => Role::ColorWell,
             WidgetType::ProgressIndicator => Role::ProgressIndicator,
+            WidgetType::Window => Role::Window,
             WidgetType::Other => Role::Unknown,
         });
         if !info.enabled {
             builder.set_disabled();
         }
         if let Some(label) = info.label {
-            builder.set_name(label);
+            if matches!(builder.role(), Role::Label) {
+                builder.set_value(label);
+            } else {
+                builder.set_label(label);
+            }
         }
         if let Some(value) = info.current_text_value {
             builder.set_value(value);
@@ -1056,6 +889,9 @@ impl Response {
             // Indeterminate state
             builder.set_toggled(Toggled::Mixed);
         }
+        if let Some(hint_text) = info.hint_text {
+            builder.set_placeholder(hint_text);
+        }
     }
 
     /// Associate a label with a control for accessibility.
@@ -1094,22 +930,22 @@ impl Response {
     /// let response = ui.add(Label::new("Right-click me!").sense(Sense::click()));
     /// response.context_menu(|ui| {
     ///     if ui.button("Close the menu").clicked() {
-    ///         ui.close_menu();
+    ///         ui.close();
     ///     }
     /// });
     /// # });
     /// ```
     ///
-    /// See also: [`Ui::menu_button`] and [`Ui::close_menu`].
+    /// See also: [`Ui::menu_button`] and [`Ui::close`].
     pub fn context_menu(&self, add_contents: impl FnOnce(&mut Ui)) -> Option<InnerResponse<()>> {
-        menu::context_menu(self, add_contents)
+        Popup::context_menu(self).show(add_contents)
     }
 
     /// Returns whether a context menu is currently open for this widget.
     ///
     /// See [`Self::context_menu`].
     pub fn context_menu_opened(&self) -> bool {
-        menu::context_menu_opened(self)
+        Popup::context_menu(self).is_open()
     }
 
     /// Draw a debug rectangle over the response displaying the response's id and whether it is
@@ -1125,9 +961,9 @@ impl Response {
     pub fn paint_debug_info(&self) {
         self.ctx.debug_painter().debug_rect(
             self.rect,
-            if self.hovered {
+            if self.hovered() {
                 crate::Color32::DARK_GREEN
-            } else if self.enabled {
+            } else if self.enabled() {
                 crate::Color32::BLUE
             } else {
                 crate::Color32::RED
@@ -1145,7 +981,10 @@ impl Response {
     ///
     /// You may not call [`Self::interact`] on the resulting `Response`.
     pub fn union(&self, other: Self) -> Self {
-        assert!(self.ctx == other.ctx);
+        assert!(
+            self.ctx == other.ctx,
+            "Responses must be from the same `Context`"
+        );
         debug_assert!(
             self.layer_id == other.layer_id,
             "It makes no sense to combine Responses from two different layers"
@@ -1157,20 +996,8 @@ impl Response {
             rect: self.rect.union(other.rect),
             interact_rect: self.interact_rect.union(other.interact_rect),
             sense: self.sense.union(other.sense),
-            enabled: self.enabled || other.enabled,
-            contains_pointer: self.contains_pointer || other.contains_pointer,
-            hovered: self.hovered || other.hovered,
-            highlighted: self.highlighted || other.highlighted,
-            clicked: self.clicked || other.clicked,
-            fake_primary_click: self.fake_primary_click || other.fake_primary_click,
-            long_touched: self.long_touched || other.long_touched,
-            drag_started: self.drag_started || other.drag_started,
-            dragged: self.dragged || other.dragged,
-            drag_stopped: self.drag_stopped || other.drag_stopped,
-            is_pointer_button_down_on: self.is_pointer_button_down_on
-                || other.is_pointer_button_down_on,
+            flags: self.flags | other.flags,
             interact_pointer_pos: self.interact_pointer_pos.or(other.interact_pointer_pos),
-            changed: self.changed || other.changed,
             intrinsic_size: None,
         }
     }
