@@ -8,6 +8,6 @@ mod label_text_selection;
 pub mod text_cursor_state;
 pub mod visuals;
 
-pub use cursor_range::{CCursorRange, CursorRange, PCursorRange};
+pub use cursor_range::CCursorRange;
 pub use label_text_selection::LabelSelectionState;
 pub use text_cursor_state::TextCursorState;
