@@ -19,9 +19,9 @@ use std::{cmp::Ordering, ops::RangeInclusive, sync::Arc};
 
 use ahash::HashMap;
 use egui::{
-    epaint, remap_clamp, vec2, Align2, Color32, CursorIcon, Id, Layout, NumExt, PointerButton,
-    Pos2, Rangef, Rect, Response, Rounding, Sense, Shape, Stroke, TextStyle, Ui, Vec2, Vec2b,
-    WidgetText,
+    Align2, Color32, CursorIcon, Id, Layout, NumExt as _, PointerButton, Pos2, Rangef, Rect,
+    Response, Sense, Shape, Stroke, TextStyle, Ui, Vec2, Vec2b, WidgetText, epaint, remap_clamp,
+    vec2,
 };
 use emath::Float as _;
 
@@ -29,10 +29,10 @@ pub use crate::{
     axis::{Axis, AxisHints, HPlacement, Placement, VPlacement},
     items::{
         Arrows, Bar, BarChart, BoxElem, BoxPlot, BoxSpread, ClosestElem, HLine, Line, LineStyle,
-        MarkerShape, Orientation, PlotConfig, PlotGeometry, PlotImage, PlotItem, PlotPoint,
-        PlotPoints, Points, Polygon, Text, VLine,
+        MarkerShape, Orientation, PlotConfig, PlotGeometry, PlotImage, PlotItem, PlotItemBase,
+        PlotPoint, PlotPoints, Points, Polygon, Text, VLine,
     },
-    legend::{Corner, Legend},
+    legend::{ColorConflictHandling, Corner, Legend},
     memory::PlotMemory,
     plot_ui::PlotUi,
     transform::{PlotBounds, PlotTransform},
@@ -126,7 +126,9 @@ pub struct PlotResponse<R> {
 
     /// The id of a currently hovered item if any.
     ///
-    /// This is `None` if either no item was hovered, or the hovered item didn't provide an id.
+    /// This is `None` if either no item was hovered.
+    /// A plot item can be hovered either by hovering its representation in the plot (line, marker, etc.)
+    /// or by hovering the item in the legend.
     pub hovered_plot_item: Option<Id>,
 }
 
@@ -144,7 +146,7 @@ pub struct PlotResponse<R> {
 ///     let x = i as f64 * 0.01;
 ///     [x, x.sin()]
 /// }).collect();
-/// let line = Line::new(sin);
+/// let line = Line::new("sin", sin);
 /// Plot::new("my_plot").view_aspect(2.0).show(ui, |plot_ui| plot_ui.line(line));
 /// # });
 /// ```
@@ -155,6 +157,7 @@ pub struct Plot<'a> {
     center_axis: Vec2b,
     allow_zoom: Vec2b,
     allow_drag: Vec2b,
+    allow_axis_zoom_drag: Vec2b,
     allow_scroll: Vec2b,
     allow_double_click_reset: bool,
     allow_boxed_zoom: bool,
@@ -180,13 +183,13 @@ pub struct Plot<'a> {
     x_axes: Vec<AxisHints<'a>>, // default x axes
     y_axes: Vec<AxisHints<'a>>, // default y axes
     legend_config: Option<Legend>,
+    cursor_color: Option<Color32>,
     show_background: bool,
     show_axes: Vec2b,
 
     show_grid: Vec2b,
     grid_spacing: Rangef,
     grid_spacers: [GridSpacer<'a>; 2],
-    sharp_grid_lines: bool,
     clamp_grid: bool,
 
     sense: Sense,
@@ -202,6 +205,7 @@ impl<'a> Plot<'a> {
             center_axis: false.into(),
             allow_zoom: true.into(),
             allow_drag: true.into(),
+            allow_axis_zoom_drag: true.into(),
             allow_scroll: true.into(),
             allow_double_click_reset: true,
             allow_boxed_zoom: true,
@@ -227,13 +231,13 @@ impl<'a> Plot<'a> {
             x_axes: vec![AxisHints::new(Axis::X)],
             y_axes: vec![AxisHints::new(Axis::Y)],
             legend_config: None,
+            cursor_color: None,
             show_background: true,
             show_axes: true.into(),
 
             show_grid: true.into(),
             grid_spacing: Rangef::new(8.0, 300.0),
             grid_spacers: [log_grid_spacer(10), log_grid_spacer(10)],
-            sharp_grid_lines: true,
             clamp_grid: false,
 
             sense: egui::Sense::click_and_drag(),
@@ -387,6 +391,16 @@ impl<'a> Plot<'a> {
         self
     }
 
+    /// Whether to allow dragging in the axis areas to zoom the plot. Default: `true`.
+    #[inline]
+    pub fn allow_axis_zoom_drag<T>(mut self, on: T) -> Self
+    where
+        T: Into<Vec2b>,
+    {
+        self.allow_axis_zoom_drag = on.into();
+        self
+    }
+
     /// Provide a function to customize the on-hover label for the x and y axis
     ///
     /// ```
@@ -396,7 +410,7 @@ impl<'a> Plot<'a> {
     ///     let x = i as f64 * 0.01;
     ///     [x, x.sin()]
     /// }).collect();
-    /// let line = Line::new(sin);
+    /// let line = Line::new("sin", sin);
     /// Plot::new("my_plot").view_aspect(2.0)
     /// .label_formatter(|name, value| {
     ///     if !name.is_empty() {
@@ -500,6 +514,38 @@ impl<'a> Plot<'a> {
         self
     }
 
+    /// Overwrite the starting and reset bounds used for the x axis.
+    /// Set the `default_auto_bounds` of the x axis to `false`.
+    ///
+    /// Panics in debug builds if `min >= max`.
+    #[inline]
+    pub fn default_x_bounds(mut self, min: f64, max: f64) -> Self {
+        debug_assert!(
+            min < max,
+            "`min` must be less than `max` in `default_x_bounds`"
+        );
+        self.default_auto_bounds.x = false;
+        self.min_auto_bounds.min[0] = min;
+        self.min_auto_bounds.max[0] = max;
+        self
+    }
+
+    /// Overwrite the starting and reset bounds used for the y axis.
+    /// Set the `default_auto_bounds` of the y axis to `false`.
+    ///
+    /// Panics in debug builds if `min >= max`.
+    #[inline]
+    pub fn default_y_bounds(mut self, min: f64, max: f64) -> Self {
+        debug_assert!(
+            min < max,
+            "`min` must be less than `max` in `default_y_bounds`"
+        );
+        self.default_auto_bounds.y = false;
+        self.min_auto_bounds.min[1] = min;
+        self.min_auto_bounds.max[1] = max;
+        self
+    }
+
     /// Expand bounds to include the given x value.
     /// For instance, to always show the y axis, call `plot.include_x(0.0)`.
     #[inline]
@@ -520,8 +566,8 @@ impl<'a> Plot<'a> {
     ///
     /// This is enabled by default.
     #[inline]
-    pub fn auto_bounds(mut self, auto_bounds: Vec2b) -> Self {
-        self.default_auto_bounds = auto_bounds;
+    pub fn auto_bounds(mut self, auto_bounds: impl Into<Vec2b>) -> Self {
+        self.default_auto_bounds = auto_bounds.into();
         self
     }
 
@@ -578,39 +624,25 @@ impl<'a> Plot<'a> {
 
     /// Add this plot to an axis link group so that this plot will share the bounds with other plots in the
     /// same group. A plot cannot belong to more than one axis group.
-    #[allow(clippy::fn_params_excessive_bools)] // TODO(emilk): use a `Vec2` as argument instead
     #[inline]
-    pub fn link_axis(mut self, group_id: impl Into<Id>, link_x: bool, link_y: bool) -> Self {
-        self.linked_axes = Some((
-            group_id.into(),
-            Vec2b {
-                x: link_x,
-                y: link_y,
-            },
-        ));
+    pub fn link_axis(mut self, group_id: impl Into<Id>, link: impl Into<Vec2b>) -> Self {
+        self.linked_axes = Some((group_id.into(), link.into()));
         self
     }
 
     /// Add this plot to a cursor link group so that this plot will share the cursor position with other plots
     /// in the same group. A plot cannot belong to more than one cursor group.
-    #[allow(clippy::fn_params_excessive_bools)] // TODO(emilk): use a `Vec2` as argument instead
     #[inline]
-    pub fn link_cursor(mut self, group_id: impl Into<Id>, link_x: bool, link_y: bool) -> Self {
-        self.linked_cursors = Some((
-            group_id.into(),
-            Vec2b {
-                x: link_x,
-                y: link_y,
-            },
-        ));
+    pub fn link_cursor(mut self, group_id: impl Into<Id>, link: impl Into<Vec2b>) -> Self {
+        self.linked_cursors = Some((group_id.into(), link.into()));
         self
     }
 
     /// Round grid positions to full pixels to avoid aliasing. Improves plot appearance but might have an
     /// undesired effect when shifting the plot bounds. Enabled by default.
     #[inline]
-    pub fn sharp_grid_lines(mut self, enabled: bool) -> Self {
-        self.sharp_grid_lines = enabled;
+    #[deprecated = "This no longer has any effect and is always enabled."]
+    pub fn sharp_grid_lines(self, _enabled: bool) -> Self {
         self
     }
 
@@ -727,21 +759,30 @@ impl<'a> Plot<'a> {
         self
     }
 
+    /// Set custom cursor color.
+    ///
+    /// You may set the color to [`Color32::TRANSPARENT`] to hide the cursors.
+    #[inline]
+    pub fn cursor_color(mut self, color: Color32) -> Self {
+        self.cursor_color = Some(color);
+        self
+    }
+
     /// Interact with and add items to the plot and finally draw it.
-    pub fn show<R>(
+    pub fn show<'b, R>(
         self,
         ui: &mut Ui,
-        build_fn: impl FnOnce(&mut PlotUi) -> R + 'a,
+        build_fn: impl FnOnce(&mut PlotUi<'b>) -> R + 'a,
     ) -> PlotResponse<R> {
         self.show_dyn(ui, Box::new(build_fn))
     }
 
     #[allow(clippy::too_many_lines)] // TODO(emilk): shorten this function
     #[allow(clippy::type_complexity)] // build_fn
-    fn show_dyn<R>(
+    fn show_dyn<'b, R>(
         self,
         ui: &mut Ui,
-        build_fn: Box<dyn FnOnce(&mut PlotUi) -> R + 'a>,
+        build_fn: Box<dyn FnOnce(&mut PlotUi<'b>) -> R + 'a>,
     ) -> PlotResponse<R> {
         let Self {
             id_source,
@@ -749,6 +790,7 @@ impl<'a> Plot<'a> {
             center_axis,
             allow_zoom,
             allow_drag,
+            allow_axis_zoom_drag,
             allow_scroll,
             allow_double_click_reset,
             allow_boxed_zoom,
@@ -768,6 +810,7 @@ impl<'a> Plot<'a> {
             x_axes,
             y_axes,
             legend_config,
+            cursor_color,
             reset,
             show_background,
             show_axes,
@@ -778,7 +821,6 @@ impl<'a> Plot<'a> {
 
             clamp_grid,
             grid_spacers,
-            sharp_grid_lines,
             sense,
         } = self;
 
@@ -835,6 +877,31 @@ impl<'a> Plot<'a> {
         // Allocate the plot window.
         let response = ui.allocate_rect(plot_rect, sense);
 
+        let x_axis_responses = x_axis_widgets
+            .iter()
+            .map(|widget| {
+                let axis_response = ui.allocate_rect(widget.rect, Sense::drag());
+                if allow_axis_zoom_drag.x {
+                    axis_response.on_hover_cursor(CursorIcon::ResizeHorizontal)
+                } else {
+                    axis_response
+                }
+            })
+            .collect::<Vec<_>>();
+
+        let y_axis_responses = y_axis_widgets
+            .iter()
+            .map(|widget| {
+                let axis_response = ui.allocate_rect(widget.rect, Sense::drag());
+
+                if allow_axis_zoom_drag.y {
+                    axis_response.on_hover_cursor(CursorIcon::ResizeVertical)
+                } else {
+                    axis_response
+                }
+            })
+            .collect::<Vec<_>>();
+
         // Load or initialize the memory.
         ui.ctx().check_for_id_clash(plot_id, plot_rect, "Plot");
 
@@ -853,7 +920,7 @@ impl<'a> Plot<'a> {
             auto_bounds: default_auto_bounds,
             hovered_legend_item: None,
             hidden_items: Default::default(),
-            transform: PlotTransform::new(plot_rect, min_auto_bounds, center_axis.x, center_axis.y),
+            transform: PlotTransform::new(plot_rect, min_auto_bounds, center_axis),
             last_click_pos_for_zoom: None,
             x_axis_thickness: Default::default(),
             y_axis_thickness: Default::default(),
@@ -886,9 +953,10 @@ impl<'a> Plot<'a> {
                 .with_clip_rect(plot_rect)
                 .add(epaint::RectShape::new(
                     plot_rect,
-                    Rounding::same(2.0),
+                    2,
                     ui.visuals().extreme_bg_color,
                     ui.visuals().widgets.noninteractive.bg_stroke,
+                    egui::StrokeKind::Inside,
                 ));
         }
 
@@ -901,12 +969,12 @@ impl<'a> Plot<'a> {
             show_y = false;
         }
         // Remove the deselected items.
-        items.retain(|item| !mem.hidden_items.contains(item.name()));
+        items.retain(|item| !mem.hidden_items.contains(&item.id()));
         // Highlight the hovered items.
-        if let Some(hovered_name) = &mem.hovered_legend_item {
+        if let Some(item_id) = &mem.hovered_legend_item {
             items
                 .iter_mut()
-                .filter(|entry| entry.name() == hovered_name)
+                .filter(|entry| &entry.id() == item_id)
                 .for_each(|entry| entry.highlight());
         }
         // Move highlighted items to front.
@@ -965,12 +1033,19 @@ impl<'a> Plot<'a> {
             mem.auto_bounds = true.into();
         }
 
+        let any_dynamic_modifications = !bounds_modifications.is_empty();
         // Apply bounds modifications.
         for modification in bounds_modifications {
             match modification {
-                BoundsModification::Set(new_bounds) => {
-                    bounds = new_bounds;
-                    mem.auto_bounds = false.into();
+                BoundsModification::SetX(range) => {
+                    bounds.min[0] = *range.start();
+                    bounds.max[0] = *range.end();
+                    mem.auto_bounds.x = false;
+                }
+                BoundsModification::SetY(range) => {
+                    bounds.min[1] = *range.start();
+                    bounds.max[1] = *range.end();
+                    mem.auto_bounds.y = false;
                 }
                 BoundsModification::Translate(delta) => {
                     let delta = (delta.x as f64, delta.y as f64);
@@ -988,10 +1063,10 @@ impl<'a> Plot<'a> {
         }
 
         // Reset bounds to initial bounds if they haven't been modified.
-        if mem.auto_bounds.x {
+        if (!default_auto_bounds.x && !any_dynamic_modifications) || mem.auto_bounds.x {
             bounds.set_x(&min_auto_bounds);
         }
-        if mem.auto_bounds.y {
+        if (!default_auto_bounds.y && !any_dynamic_modifications) || mem.auto_bounds.y {
             bounds.set_y(&min_auto_bounds);
         }
 
@@ -1019,7 +1094,7 @@ impl<'a> Plot<'a> {
             }
         }
 
-        mem.transform = PlotTransform::new(plot_rect, bounds, center_axis.x, center_axis.y);
+        mem.transform = PlotTransform::new(plot_rect, bounds, center_axis);
 
         // Enforce aspect ratio
         if let Some(data_aspect) = data_aspect {
@@ -1052,6 +1127,42 @@ impl<'a> Plot<'a> {
             mem.auto_bounds = mem.auto_bounds.and(!allow_drag);
         }
 
+        // Drag axes to zoom:
+        for d in 0..2 {
+            if allow_axis_zoom_drag[d] {
+                if let Some(axis_response) = (if d == 0 {
+                    &x_axis_responses
+                } else {
+                    &y_axis_responses
+                })
+                .iter()
+                .find(|r| r.dragged_by(PointerButton::Primary))
+                {
+                    if let Some(drag_start_pos) = ui.input(|i| i.pointer.press_origin()) {
+                        let delta = axis_response.drag_delta();
+
+                        let axis_zoom = 1.0 + (0.02 * delta[d]).clamp(-1.0, 1.0);
+
+                        let zoom = if data_aspect.is_some() {
+                            // Zoom both axes equally to maintain aspect ratio:
+                            Vec2::splat(axis_zoom)
+                        } else {
+                            let mut zoom = Vec2::splat(1.0);
+                            zoom[d] = axis_zoom;
+                            zoom
+                        };
+
+                        if zoom != Vec2::splat(1.0) {
+                            let mut zoom_center = plot_rect.center();
+                            zoom_center[d] = drag_start_pos[d];
+                            mem.transform.zoom(zoom, zoom_center);
+                            mem.auto_bounds = false.into();
+                        }
+                    }
+                }
+            }
+        }
+
         // Zooming
         let mut boxed_zoom_rect = None;
         if allow_boxed_zoom {
@@ -1072,11 +1183,13 @@ impl<'a> Plot<'a> {
                             rect,
                             0.0,
                             epaint::Stroke::new(4., Color32::DARK_BLUE),
+                            egui::StrokeKind::Middle,
                         ), // Outer stroke
                         epaint::RectShape::stroke(
                             rect,
                             0.0,
                             epaint::Stroke::new(2., Color32::WHITE),
+                            egui::StrokeKind::Middle,
                         ), // Inner stroke
                     ));
                 }
@@ -1108,7 +1221,7 @@ impl<'a> Plot<'a> {
         // For instance: The user is painting another interactive widget on top of the plot
         // but they still want to be able to pan/zoom the plot.
         if let (true, Some(hover_pos)) = (
-            response.contains_pointer,
+            response.contains_pointer(),
             ui.input(|i| i.pointer.hover_pos()),
         ) {
             if allow_zoom.any() {
@@ -1185,6 +1298,7 @@ impl<'a> Plot<'a> {
         }
 
         let prepared = PreparedPlot {
+            plot_area_response: &response,
             items,
             show_x,
             show_y,
@@ -1193,15 +1307,15 @@ impl<'a> Plot<'a> {
             show_grid,
             grid_spacing,
             transform: mem.transform,
-            draw_cursor_x: linked_cursors.as_ref().map_or(false, |group| group.1.x),
-            draw_cursor_y: linked_cursors.as_ref().map_or(false, |group| group.1.y),
+            draw_cursor_x: linked_cursors.as_ref().is_some_and(|group| group.1.x),
+            draw_cursor_y: linked_cursors.as_ref().is_some_and(|group| group.1.y),
             draw_cursors,
+            cursor_color,
             grid_spacers,
-            sharp_grid_lines,
             clamp_grid,
         };
 
-        let (plot_cursors, hovered_plot_item) = prepared.ui(ui, &response);
+        let (plot_cursors, mut hovered_plot_item) = prepared.ui(ui, &response);
 
         if let Some(boxed_zoom_rect) = boxed_zoom_rect {
             ui.painter()
@@ -1215,7 +1329,11 @@ impl<'a> Plot<'a> {
         if let Some(mut legend) = legend {
             ui.add(&mut legend);
             mem.hidden_items = legend.hidden_items();
-            mem.hovered_legend_item = legend.hovered_item_name();
+            mem.hovered_legend_item = legend.hovered_item();
+
+            if let Some(item_id) = &mem.hovered_legend_item {
+                hovered_plot_item.get_or_insert(*item_id);
+            }
         }
 
         if let Some((id, _)) = linked_cursors.as_ref() {
@@ -1267,7 +1385,7 @@ impl<'a> Plot<'a> {
 /// Returns the rect left after adding axes.
 fn axis_widgets<'a>(
     mem: Option<&PlotMemory>,
-    show_axes: Vec2b,
+    show_axes: impl Into<Vec2b>,
     complete_rect: Rect,
     [x_axes, y_axes]: [&'a [AxisHints<'a>]; 2],
 ) -> ([Vec<AxisWidget<'a>>; 2], Rect) {
@@ -1293,6 +1411,7 @@ fn axis_widgets<'a>(
     //      |      X-axis 1      |   |
     //  +   +--------------------+---+
     //
+    let show_axes = show_axes.into();
 
     let mut x_axis_widgets = Vec::<AxisWidget<'_>>::new();
     let mut y_axis_widgets = Vec::<AxisWidget<'_>>::new();
@@ -1305,7 +1424,7 @@ fn axis_widgets<'a>(
         let initial_x_range = complete_rect.x_range();
 
         for (i, cfg) in x_axes.iter().enumerate().rev() {
-            let mut height = cfg.thickness(Axis::X);
+            let mut height = cfg.min_thickness;
             if let Some(mem) = mem {
                 // If the labels took up too much space the previous frame, give them more space now:
                 height = height.max(mem.x_axis_thickness.get(&i).copied().unwrap_or_default());
@@ -1333,7 +1452,7 @@ fn axis_widgets<'a>(
         let plot_y_range = rect_left.y_range();
 
         for (i, cfg) in y_axes.iter().enumerate().rev() {
-            let mut width = cfg.thickness(Axis::Y);
+            let mut width = cfg.min_thickness;
             if let Some(mem) = mem {
                 // If the labels took up too much space the previous frame, give them more space now:
                 width = width.max(mem.y_axis_thickness.get(&i).copied().unwrap_or_default());
@@ -1357,6 +1476,12 @@ fn axis_widgets<'a>(
         }
     }
 
+    // The loops iterated through {x,y}_axes in reverse order, so we have to reverse the
+    // {x,y}_axis_widgets vec as well. Otherwise, the indices are messed up and the plot memory
+    // (mem.{x,y}_axis_thickness) will access the wrong axis given an index.
+    x_axis_widgets.reverse();
+    y_axis_widgets.reverse();
+
     let mut plot_rect = rect_left;
 
     // If too little space, remove axis widgets
@@ -1378,7 +1503,8 @@ fn axis_widgets<'a>(
 /// User-requested modifications to the plot bounds. We collect them in the plot build function to later apply
 /// them at the right time, as other modifications need to happen first.
 enum BoundsModification {
-    Set(PlotBounds),
+    SetX(RangeInclusive<f64>),
+    SetY(RangeInclusive<f64>),
     Translate(Vec2),
     AutoBounds(Vec2b),
     Zoom(Vec2, PlotPoint),
@@ -1467,7 +1593,9 @@ pub fn uniform_grid_spacer<'a>(spacer: impl Fn(GridInput) -> [f64; 3] + 'a) -> G
 // ----------------------------------------------------------------------------
 
 struct PreparedPlot<'a> {
-    items: Vec<Box<dyn PlotItem>>,
+    /// The response of the whole plot area
+    plot_area_response: &'a Response,
+    items: Vec<Box<dyn PlotItem + 'a>>,
     show_x: bool,
     show_y: bool,
     label_formatter: LabelFormatter<'a>,
@@ -1480,12 +1608,12 @@ struct PreparedPlot<'a> {
     draw_cursor_x: bool,
     draw_cursor_y: bool,
     draw_cursors: Vec<Cursor>,
+    cursor_color: Option<Color32>,
 
-    sharp_grid_lines: bool,
     clamp_grid: bool,
 }
 
-impl<'a> PreparedPlot<'a> {
+impl PreparedPlot<'_> {
     fn ui(self, ui: &mut Ui, response: &Response) -> (Vec<Cursor>, Option<Id>) {
         let mut axes_shapes = Vec::new();
 
@@ -1521,7 +1649,7 @@ impl<'a> PreparedPlot<'a> {
         };
 
         // Draw cursors
-        let line_color = rulers_color(ui);
+        let line_color = self.cursor_color.unwrap_or_else(|| rulers_color(ui));
 
         let mut draw_cursor = |cursors: &Vec<Cursor>, always| {
             for &cursor in cursors {
@@ -1554,6 +1682,7 @@ impl<'a> PreparedPlot<'a> {
         let painter = ui.painter().with_clip_rect(*transform.frame());
         painter.extend(shapes);
 
+        // Show coordinates in a corner of the plot:
         if let Some((corner, formatter)) = self.coordinates_formatter.as_ref() {
             let hover_pos = response.hover_pos();
             if let Some(pointer) = hover_pos {
@@ -1567,7 +1696,16 @@ impl<'a> PreparedPlot<'a> {
                     Corner::LeftBottom => (Align2::LEFT_BOTTOM, padded_frame.left_bottom()),
                     Corner::RightBottom => (Align2::RIGHT_BOTTOM, padded_frame.right_bottom()),
                 };
-                painter.text(position, anchor, text, font_id, ui.visuals().text_color());
+
+                let text_color = ui.visuals().text_color();
+                let galley = painter.layout_no_wrap(text, font_id, text_color);
+                let rect = anchor.anchor_size(position, galley.size());
+                painter.rect_filled(
+                    rect.expand(4.0),
+                    ui.style().visuals.window_corner_radius,
+                    ui.style().visuals.extreme_bg_color.gamma_multiply(0.75),
+                );
+                painter.galley(rect.min, galley, text_color);
             }
         }
 
@@ -1658,12 +1796,6 @@ impl<'a> PreparedPlot<'a> {
                 }
             }
 
-            if self.sharp_grid_lines {
-                // Round to avoid aliasing
-                p0 = ui.painter().round_pos_to_pixels(p0);
-                p1 = ui.painter().round_pos_to_pixels(p1);
-            }
-
             shapes.push((
                 Shape::line_segment([p0, p1], Stroke::new(1.0, line_color)),
                 line_strength,
@@ -1673,6 +1805,7 @@ impl<'a> PreparedPlot<'a> {
 
     fn hover(&self, ui: &Ui, pointer: Pos2, shapes: &mut Vec<Shape>) -> (Vec<Cursor>, Option<Id>) {
         let Self {
+            plot_area_response,
             transform,
             show_x,
             show_y,
@@ -1711,16 +1844,22 @@ impl<'a> PreparedPlot<'a> {
         let mut cursors = Vec::new();
 
         let hovered_plot_item_id = if let Some((item, elem)) = closest {
-            item.on_hover(elem, shapes, &mut cursors, &plot, label_formatter);
-            item.id()
+            item.on_hover(
+                plot_area_response,
+                elem,
+                shapes,
+                &mut cursors,
+                &plot,
+                label_formatter,
+            );
+            Some(item.id())
         } else {
             let value = transform.value_from_position(pointer);
-            items::rulers_at_value(
-                pointer,
+            items::rulers_and_tooltip_at_value(
+                plot_area_response,
                 value,
                 "",
                 &plot,
-                shapes,
                 &mut cursors,
                 label_formatter,
             );
