@@ -1,12 +1,12 @@
 use std::collections::BTreeMap;
 use std::sync::Arc;
 
-use emath::{vec2, Vec2};
+use emath::{GuiRounding as _, Vec2, vec2};
 
 use crate::{
+    TextureAtlas,
     mutex::{Mutex, RwLock},
     text::FontTweak,
-    TextureAtlas,
 };
 
 // ----------------------------------------------------------------------------
@@ -91,27 +91,29 @@ impl FontImpl {
         scale_in_pixels: f32,
         tweak: FontTweak,
     ) -> Self {
-        assert!(scale_in_pixels > 0.0);
-        assert!(pixels_per_point > 0.0);
+        assert!(
+            scale_in_pixels > 0.0,
+            "scale_in_pixels is smaller than 0, got: {scale_in_pixels:?}"
+        );
+        assert!(
+            pixels_per_point > 0.0,
+            "pixels_per_point must be greater than 0, got: {pixels_per_point:?}"
+        );
 
-        use ab_glyph::{Font, ScaleFont};
+        use ab_glyph::{Font as _, ScaleFont as _};
         let scaled = ab_glyph_font.as_scaled(scale_in_pixels);
-        let ascent = scaled.ascent() / pixels_per_point;
-        let descent = scaled.descent() / pixels_per_point;
-        let line_gap = scaled.line_gap() / pixels_per_point;
+        let ascent = (scaled.ascent() / pixels_per_point).round_ui();
+        let descent = (scaled.descent() / pixels_per_point).round_ui();
+        let line_gap = (scaled.line_gap() / pixels_per_point).round_ui();
 
         // Tweak the scale as the user desired
         let scale_in_pixels = scale_in_pixels * tweak.scale;
+        let scale_in_points = scale_in_pixels / pixels_per_point;
 
-        let baseline_offset = {
-            let scale_in_points = scale_in_pixels / pixels_per_point;
-            scale_in_points * tweak.baseline_offset_factor
-        };
+        let baseline_offset = (scale_in_points * tweak.baseline_offset_factor).round_ui();
 
-        let y_offset_points = {
-            let scale_in_points = scale_in_pixels / pixels_per_point;
-            scale_in_points * tweak.y_offset_factor
-        } + tweak.y_offset;
+        let y_offset_points =
+            ((scale_in_points * tweak.y_offset_factor) + tweak.y_offset).round_ui();
 
         // Center scaled glyphs properly:
         let height = ascent + descent;
@@ -147,14 +149,6 @@ impl FontImpl {
             return false;
         }
 
-        if self.name == "emoji-icon-font" {
-            // HACK: https://github.com/emilk/egui/issues/1284 https://github.com/jslegers/emoji-icon-font/issues/18
-            // Don't show the wrong fullwidth capital letters:
-            if 'Ｓ' <= chr && chr <= 'Ｙ' {
-                return true;
-            }
-        }
-
         matches!(
             chr,
             // Strip out a religious symbol with secondary nefarious interpretation:
@@ -239,7 +233,7 @@ impl FontImpl {
         last_glyph_id: ab_glyph::GlyphId,
         glyph_id: ab_glyph::GlyphId,
     ) -> f32 {
-        use ab_glyph::{Font as _, ScaleFont};
+        use ab_glyph::{Font as _, ScaleFont as _};
         self.ab_glyph_font
             .as_scaled(self.scale_in_pixels as f32)
             .kern(last_glyph_id, glyph_id)
@@ -247,6 +241,8 @@ impl FontImpl {
     }
 
     /// Height of one row of text in points.
+    ///
+    /// Returns a value rounded to [`emath::GUI_ROUNDING`].
     #[inline(always)]
     pub fn row_height(&self) -> f32 {
         self.height_in_points
@@ -266,8 +262,8 @@ impl FontImpl {
     }
 
     fn allocate_glyph(&self, glyph_id: ab_glyph::GlyphId) -> GlyphInfo {
-        assert!(glyph_id.0 != 0);
-        use ab_glyph::{Font as _, ScaleFont};
+        assert!(glyph_id.0 != 0, "Can't allocate glyph for id 0");
+        use ab_glyph::{Font as _, ScaleFont as _};
 
         let glyph = glyph_id.with_scale_and_position(
             self.scale_in_pixels as f32,
@@ -283,12 +279,13 @@ impl FontImpl {
             } else {
                 let glyph_pos = {
                     let atlas = &mut self.atlas.lock();
+                    let text_alpha_from_coverage = atlas.text_alpha_from_coverage;
                     let (glyph_pos, image) = atlas.allocate((glyph_width, glyph_height));
                     glyph.draw(|x, y, v| {
                         if 0.0 < v {
                             let px = glyph_pos.0 + x as usize;
                             let py = glyph_pos.1 + y as usize;
-                            image[(px, py)] = v;
+                            image[(px, py)] = text_alpha_from_coverage.color_from_coverage(v);
                         }
                     });
                     glyph_pos
@@ -418,7 +415,9 @@ impl Font {
         (point * self.pixels_per_point).round() / self.pixels_per_point
     }
 
-    /// Height of one row of text. In points
+    /// Height of one row of text. In points.
+    ///
+    /// Returns a value rounded to [`emath::GUI_ROUNDING`].
     #[inline(always)]
     pub fn row_height(&self) -> f32 {
         self.row_height
